import numpy as np
import tohil


#-----------------------------------------------------------------------------------------------------------------------------------------
#-----------------------------------------------------------------------------------------------------------------------------------------
def createInputAbaqus(out_file_name, tensileSprings):
   
    tensileSprings = int(tensileSprings)

    # Leemos variables de malla
    nodes, elements, joins, outer_ids = getMeshVar()
    # Creamos un array con tensile connectors if covered
    if tensileSprings == 1:
        tens_list = tensileSpringFunction(nodes, outer_ids)
    else:
        tens_list = []
    # Escribimos el Abaqus input file
    writeAbaqusInput(out_file_name, elements, nodes, joins, tens_list, tensileSprings)
    return 0


#-----------------------------------------------------------------------------------------------------------------------------------------
#-----------------------------------------------------------------------------------------------------------------------------------------
def computeDist(p1,p2):
    try:
        d = 0
        for i in range(len(p1)):
            d = d +  (p1[i] - p2[i])**2
        d = np.sqrt(d)
        return d
    except:
        raise('Error distance function')
    
#-----------------------------------------------------------------------------------------------------------------  
def cart2pol(x, y):
    rho = np.sqrt(x ** 2 + y ** 2)
    phi = np.arctan2(y, x)
    return (rho, phi)

#-----------------------------------------------------------------------------------------------------------------
def getMeshVar():
    #Read mesh nodes and elements    
    tcl = tohil.import_tcl()

    nodes1 = tcl.GiD_Info('mesh','nodes')
    outer_nodes1 = tcl.GiD_Info('mesh','nodes', '-group','outer nodes') 
    top = tcl.GiD_Info('mesh','nodes', '-group','top') 
    bottom = tcl.GiD_Info('mesh','nodes', '-group','bottom')
    elements_Abqs1 = tcl.GiD_Info('mesh','elements','Linear', '-group','wires_1')
    elements_Abqs2 = tcl.GiD_Info('mesh','elements','Linear', '-group','wires_2')
    wires = tcl.GiD_Info('mesh','elements','Linear', '-group','joints')

    #Creamos un array con las conectividades
    m = len(nodes1)
    nodes = [[float(nodes1[i+1]), float(nodes1[i+2]), float(nodes1[i+3])] for i in range(0, m, 4) if i+3 < m]

    # Creamos un array con elementos
    n1 = len(elements_Abqs1)
    n2 = len(elements_Abqs2)
    array_w1 = [[int(elements_Abqs1[i+1]), int(elements_Abqs1[i+2])] for i in range(0, n1, 4) if i+2 < n1]
    array_w2 = [[int(elements_Abqs2[i+1]), int(elements_Abqs2[i+2])] for i in range(0, n2, 4) if i+2 < n2]
    elements = array_w1 + array_w2

    # Creamos un array con wires connectors
    nc = len(wires)
    joins = [[int(wires[i+1]), int(wires[i+2])] for i in range(0, nc, 4) if i+2 < nc]

    # Creamos lista de outer ids
    mo = len(outer_nodes1)
    outer_ids = [int(outer_nodes1[i]) for i in range(0, mo, 4) if i < mo] 
    top_ids = [int(top[i]) for i in range(0, len(top), 4) if i < len(top)] 
    bottom_ids = [int(bottom[i]) for i in range(0, len(bottom), 4) if i < len(bottom)] 

    outer_ids = top_ids + outer_ids + bottom_ids

    return nodes, elements, joins, outer_ids



#-----------------------------------------------------------------------------------------------------------------
def tensileSpringFunction(nodes, outer_ids):
    #Cambio a coordenadas cilindricas
    list_coor = []
    for i in outer_ids:
        [rho, phi] = cart2pol(nodes[i-1][0], nodes[i-1][2])
        list_coor.append([rho, nodes[i-1][1], phi]) 

    Matrix_pol1 = np.array(list_coor)

    #Ordenamos las coordenadas de acuerdo al eje y
    tol_phi = 0.05 #Depends on mesh
    tol_y = 0.01
    Matrix_pol = Matrix_pol1[np.argsort(Matrix_pol1[:, 1])]
    ynn = np.unique(Matrix_pol[:, 1])
    y2 = np.where(np.diff(ynn) > tol_y)[0][0] 
    y_uniques = [ynn[0], ynn[y2 + 1]]
    tens = []
    for j in range(0,2):
        #Creamos los tensile connectors en cada eje axial de phi_unico
        #y_min = Matrix_pol[0][1]  
        y_min = y_uniques[j]
        #Creamos los tensile connectors en cada eje axial de phi_unico
        phi_unicos = set(node[2] for node in Matrix_pol if abs(node[1]-y_min) <= tol_y)
        for phi in phi_unicos:
            nodes_phi = [node for node in Matrix_pol if abs(node[2] - phi) <= tol_phi]
            nodes_phi.sort(key=lambda x: x[1])   # Ordenar los nodos por coordenada y
            for i in range(len(nodes_phi) - 1):
                y_actual, phi_actual, rho_actual = nodes_phi[i][1], nodes_phi[i][2], nodes_phi[i][0]
                y_siguiente, phi_siguiente, rho_siguiente = nodes_phi[i + 1][1],nodes_phi[i + 1][2],nodes_phi[i + 1][0]
                indice1 = np.where((Matrix_pol1[:, 2] == phi_actual) & (Matrix_pol1[:, 1] == y_actual) & (Matrix_pol1[:, 0] == rho_actual))
                indice2 = np.where((Matrix_pol1[:, 2] == phi_siguiente) & (Matrix_pol1[:, 1] == y_siguiente) & (Matrix_pol1[:, 0] == rho_siguiente))
                tens.append([outer_ids[indice1[0][0]], outer_ids[indice2[0][0]]])
    
    return tens


#-----------------------------------------------------------------------------------------------------------------
def writeAbaqusInput(out_file_name, elements, nodes, joins, tens, tensileSprings):

    out_file = open(out_file_name, "w")

    head = ['*Heading', '** Job name: STENT Model name: STENT_JOINS','** Generated by GiD - Wires stent generator','*Preprint, echo=NO, model=NO, history=NO, contact=NO']

    for word in head:
        out_file.write(word + "\n")

    print('**' + '\n'+ '** PARTS' + '\n' + '**' , file=out_file)
    print('*Part, name=STENT' + '\n' + '*Node', file=out_file)

    i = 0
    for coor in nodes:
        i += 1
        out_file.write(str(i) + ', ' + ", ".join(str(e) for e in coor) + "\n")

    out_file.write('*Element, type=B31'+ "\n")

    i = 0
    for elem in elements:
        i += 1
        out_file.write(str(i) + ', ' + ", ".join(str(e) for e in elem) + "\n")

    #ELEMENT SETS
    out_file.write('*Elset, elset=ES_PID_T_1, generate'+ "\n")
    out_file.write(str(1) + ', ' + str(len(elements)) + ', ' + str(1)+ "\n")

    out_file.write('*Elset, elset=_STENT-SURF1_, internal, generate'+ "\n")
    out_file.write(str(1) + ', ' + str(len(elements)) + ', ' + str(1)+ "\n")

    #SURFACE SETS
    out_file.write('*Surface, type=ELEMENT, name=STENT-SURF1'+ "\n")
    out_file.write('_STENT-SURF1_,'+ "\n")

    #BEAM SECTION
    out_file.write('** Section: Section-2-ES_PID_T_1  Profile: Profile-1'+ "\n")
    out_file.write('*Beam Section, elset=ES_PID_T_1, material=NITI, temperature=GRADIENTS, section=CIRC'+ "\n")
    out_file.write(str(round(computeDist(nodes[joins[0][0] - 1], nodes[joins[0][1] -1]) / 2, 3)) + "\n" + '0.,0.,-1.'+ "\n" + '*End Part'+ "\n" + '**'+ "\n" + '**'+ "\n")

    #ASSEMBLY
    out_file.write('** ASSEMBLY' + "\n" + '**'+ '\n' + '*Assembly, name=Assembly'+ '\n'+ '**'+'\n')
    lst2 = [item[1] for item in nodes]
    out_file.write('*Instance, name=STENT, part=STENT' + '\n' + '0., '  + str(-1* min(lst2)) + ', 0.' +'\n' + '*End Instance' + '\n'+ '**'  +'\n')


    #JOINS: SPRING ELEMENTS--------------------------------------------------------------------------------
    out_file.write( '*Element, type=CONN3D2' + '\n')
    ii = 0
    for elem in joins:
        ii += 1
        out_file.write(str(ii) + ', ' + 'STENT.'+ str(elem[0]) + ', ' + 'STENT.'+ str(elem[1])  + "\n")
        

    # out_file.write('*Connector Section, elset=SetRot, behavior=ConnProp-1' + '\n')
    # out_file.write('Join, Rotation' + '\n' + '"Datum csys-3",' + '\n')
    out_file.write('*Connector Section, elset=SetRot' + '\n')
    out_file.write('Join' + '\n' + '"Datum csys-3",' + '\n')


    #DEFINE TENSILE SPRINGS IF COVERED--------------------------------------------------------------------
    if tensileSprings == 1:
        out_file.write('*Element, type=CONN3D2' + '\n')
        for elem in tens:
            ii += 1
            out_file.write(str(ii) + ', ' + 'STENT.'+ str(elem[0]) + ', ' + 'STENT.'+ str(elem[1])  + "\n")
        out_file.write('*Connector Section, elset=Set-ten' + '\n' + 'Axial,' + '\n')


    #ASSEMBLY SETS---------------------------------------------------------------------------------------
    out_file.write('*Elset, elset=ALL_MASSIVE_ELEMENTS, instance=STENT, generate' + '\n')
    out_file.write(str(1) + ', ' + str(len(elements)) + ', ' + str(1)+ "\n")
    out_file.write('*Elset, elset=ALL_T3D2, instance=STENT, generate' + '\n')
    out_file.write(str(1) + ', ' + str(len(elements)) + ', ' + str(1)+ "\n")

    out_file.write('*Nset, nset=SET-3, instance=STENT, generate' + '\n')
    out_file.write(str(1) + ', ' + str(len(nodes)) + ', ' + str(1) + "\n")

    out_file.write('*Elset, elset=SetRot, internal, generate' + '\n')
    out_file.write(str(1) + ', ' + str(len(joins)) + ', ' + str(1)+ "\n")

    if tensileSprings == 1:
        out_file.write('*Elset, elset=Set-ten, internal, generate' + '\n')
        out_file.write(str(len(joins) + 1) + ', ' + str(ii) + ', ' + str(1)+ "\n")

    out_file.write('*Orientation, name="Datum csys-3", system=CYLINDRICAL' + '\n')
    out_file.write('0.,      0.,           0.,           0.,    1.,     0.' + '\n' + '1, 0.' +'\n')

    out_file.write('*End Assembly'+ '\n')

    # out_file.write('*Connector Behavior, name=ConnProp-1'+ '\n' + '*Connector Elasticity, nonlinear, component=4' + '\n' )

    #MATERIALS BEHAVIOR
    out_file.write('** '+ '\n''** MATERIALS '+ '\n''** '+ '\n' + '*Material, name=NITI' + '\n')
    out_file.write('*SUPERELASTIC' + '\n' + ' 25600.,  0.33, 0.046,  590.,  600.,  420.,  380.,    0. '+ '\n'+ ' 15.,   6.7,   6.7,    0.' +  '\n')
    out_file.write('*Elastic' + '\n' + ' 56700., 0.33 '+ '\n')

    #INTERACTION PROPERTIES
    out_file.write('** '+ '\n''** INTERACTION PROPERTIES '+ '\n''** '+ '\n')
    out_file.write('*Surface Interaction, name=FRICTION' + '\n' + '0.05,' + '\n'+'*Friction, slip tolerance=0.005' +'\n' + '0.2,' +'\n')
    out_file.write('*Surface Behavior, pressure-overclosure=HARD' + '\n')

    out_file.close()



