#Calculix read plugin: news
#v 1.0 : first version, import mesh and results .frd Calculix format
#v 2.0 : import geometry .fbd Calculix format

namespace eval Calculix {
    variable info_plugin ;#information read from .xml plugin file
    #internal use variables
    variable GidPostElementType ;#map calculix element type number to GiD post elements
    variable GidPostElementNNode ;#map calculix element type number to GiD post element number of nodes
    variable GidPostGaussPoint1 ;#map calculix element type number to GiD post predefined internal 1 gauss point name
    variable percent_done
    variable stop
    variable _merge  1 ;#to set in file browser options if want to merge or delete old model+results
    #

    set info_plugin(path) [file dirname [info script]] ;#to know where to find the files
    if { [info exists info_plugin(Version)] } {
        set other_plugin_version $info_plugin(Version)
    }
    array set info_plugin [ReadProblemtypeXml [file join $info_plugin(path) [file tail $info_plugin(path)].xml] InfoPlugin {Name Version MinimumGiDVersion}]
    if { [info exists other_plugin_version] && $other_plugin_version != $info_plugin(Version)} {
        WarnWinText [_ "There are two versions of %s: the %s and %s inside the plugins folder. One must be deleted" $info_plugin(Name) $other_plugin_version $info_plugin(Version)]
    }
    if { [GidUtils::VersionCmp $info_plugin(MinimumGiDVersion)] < 0 } {
        WarnWinText [_ "This plugin requires GiD %s or later" $info_plugin(MinimumGiDVersion)]
    }
}


proc Calculix::Init { } {
    Calculix::InitReadPost
}


#########################################################################################
######################################### READ PRE ######################################
#########################################################################################

proc Calculix::GetArcData { gid_p1 gid_p2 gid_p_center } {
    set xyz1 [lrange [GiD_Geometry get point $gid_p1] 1 3]
    set xyz2 [lrange [GiD_Geometry get point $gid_p2] 1 3]
    set xyzcenter [lrange [GiD_Geometry get point $gid_p_center] 1 3]
    set vx [MathUtils::VectorDiff $xyzcenter $xyz1]
    set radius [MathUtils::VectorModulus $vx]
    set vc2 [MathUtils::VectorDiff $xyzcenter $xyz2]
    #set radius [MathUtils::VectorModulus $vc2]
    set angle1_rad 0
    set angle2_rad [MathUtils::VectorAngle $xyzcenter $xyz1 $xyz2]
    set vz [MathUtils::VectorVectorialProd $vx $vc2]
    set vz [MathUtils::VectorNormalized $vz]
    set vx [MathUtils::VectorNormalized $vx]
    set vy [MathUtils::VectorVectorialProd $vz $vx]
    set mat4x4 [concat $vx 0 $vy 0 $vz 0 $xyzcenter 1]
    return [list {0 0} $radius $angle1_rad $angle2_rad $mat4x4]
}

#not exists this command:
#GiD_Geometry create line <num>|append polyline layer inipoint endpoint $data
#implement an alternative using a process command. only lines parameter are used
proc Calculix::CreatePolyline { layer gid_p1 gid_p2 gid_lines_and_versos } {
    set last_id_start [GiD_Info Geometry MaxNumLines]
    set gid_lines [list]
    foreach item $gid_lines_and_versos {
        lappend gid_lines  [lindex $item 0]
    }
    GiD_Process Mescape Geometry Create Polyline {*}$gid_lines escape escape
    set last_id_end [GiD_Info Geometry MaxNumLines]
    if { $last_id_end > $last_id_start } {
        set id $last_id_end
    } else {
        set id 0
        W "fail create polyline with lines: $gid_lines"
    }
    return $id
}

proc Calculix::CreateNurbsSurfaceByBoundary { layer gid_lines_and_versos } {
    set last_id_start [GiD_Info Geometry MaxNumSurfaces]
    set gid_lines [list]
    foreach item $gid_lines_and_versos {
        lappend gid_lines [lindex $item 0]
    }
    GiD_Process Mescape Geometry Create NurbsSurface {*}$gid_lines escape escape
    set last_id_end [GiD_Info Geometry MaxNumSurfaces]
    if { $last_id_end > $last_id_start } {
        set id $last_id_end
        #it has been created only with lines, without use versos, check that it has the right orientation
        lassign [lindex [GiD_Geometry get surface $id] 9] line_test verso_test
        set pos [lsearch -index 0 $gid_lines_and_versos $line_test]
        if { [lindex $gid_lines_and_versos $pos 1] != $verso_test } {
            #swap its normal
            GiD_Process Mescape Utilities SwapNormals Surfaces Select $id escape escape
        }
    } else {
        set id 0
        #W "fail create Nurbs surface with lines: $gid_lines"
    }
    return $id
}

#remove ! and % of automatic names
proc Calculix::FixAutomaticName { name } {
    set letter [string index $name 0]
    if { $letter == "!" ||  $letter == "%"} {
        set name [string range $name 1 end]
    }
    return $name
}

proc Calculix::NormalizeKnots { knots } {
    set a [lindex $knots 0]
    set b [lindex $knots end]
    set normalized_knots [list]
    foreach knot $knots {
        lappend normalized_knots [expr {double($knot-$a)/double($b-$a)}]
    }
    return $normalized_knots
}

#read .fbd files
proc Calculix::ReadPreGeometry { filename } {
    package require struct::matrix
    variable percent_done
    set fail 0
    if { ![file exists $filename] } {
        WarnWin [_ "File %s not found" $filename]
        return 1
    }
    set fp [open $filename r]
    if { $fp == "" } {
        WarnWin [_ "Cannot open file '%s'" $filename]
        return 1
    }
    #fill some local array to facilitate access
    set gid_sense(+) 0
    set gid_sense(-) 1
    set gid_swap_sense(0) 1
    set gid_swap_sense(1) 0

    set gid_entity_over(p) points
    set gid_entity_over(l) lines
    set gid_entity_over(s) surfaces
    set gid_entity_over(b) volumes
    set gid_entity_over(n) nodes
    set gid_entity_over(e) elements
    set gid_entity_over(f) faces

    set gid_array_over(p) gid_point
    set gid_array_over(l) gid_line
    set gid_array_over(s) gid_surface
    set gid_array_over(b) gid_volume
    set gid_array_over(n) gid_node
    set gid_array_over(e) gid_element
    set gid_array_over(f) gid_face

    set count(NURL) 0
    set count(NURS) 0
    set last_point_start [GiD_Info Geometry MaxNumPoints]

    set layer [GiD_Info Project LayerToUse]
    while { ![eof $fp] } {
        gets $fp line
        set key [string toupper [lindex $line 0]]
        if { $key == "**" || $key == "" } continue
        if { $key == "COMP" } {
            set name [Calculix::FixAutomaticName [lindex $line 1]]
            set operation [string tolower [string index [lindex $line 2] 0]]
            if { ![GiD_Groups exists $name] } {
                GiD_Groups create $name
            }
            if { $operation == "d" } {
                #down: must assign to group all its lower entities
            } elseif { $operation == "u" } {
                #up: must assign to group all its higher entities
            } elseif { $operation == "e" } {
                #edges: must assign to group the end points of the lines
            } else {
                #W "Unexpected operation. $line"
            }
        } elseif { $key == "CSEL" } {
            #undocummented
        } elseif { $key == "ELTY" } {
            #e.g. ELTY all HE20
            if { [string tolower [lindex $line 1]] == "all" } {
                set selection 1:end
                set element [lindex $line 2]
                set element_type [string tolower [string range $element 0 1]]
                set element_nnode [string index $element 2]
                if { $element_type == "tr" } {
                    GiD_Process Mescape Meshing ElemType Triangle {*}$selection escape escape
                    if { $element_nnode == 3 } {
                        GiD_Set Model(QuadraticType) 0
                    } else {
                        GiD_Set Model(QuadraticType) 1
                    }
                } elseif { $element_type == "qu" } {
                    GiD_Process Mescape Meshing ElemType Quadrilateral {*}$selection escape escape
                    #GiD_Process Mescape Meshing Structured Surfaces {*}$selection escape escape
                    if { $element_nnode == 4 } {
                        GiD_Set Model(QuadraticType) 0
                    } else {
                        GiD_Set Model(QuadraticType) 1
                    }
                } elseif { $element_type == "he" } {
                    GiD_Process Mescape Meshing ElemType Hexahedra {*}$selection escape escape
                    #GiD_Process Mescape Meshing Structured Volumes {*}$selection escape escape
                    if { $element_nnode == 8 } {
                        GiD_Set Model(QuadraticType) 0
                    } else {
                        GiD_Set Model(QuadraticType) 1
                    }
                } else {
                    W "unexpected element type. $line"
                }
            }
        } elseif { $key == "EXIT" } {
            #
        } elseif { $key == "GBOD" } {
            set name [Calculix::FixAutomaticName [lindex $line 1]]
            #[lindex $line 1] == NORM , not used
            set boundary_surfaces [list]
            foreach {vi si} [lrange $line 3 end] {
                set si [Calculix::FixAutomaticName $si]
                lappend boundary_surfaces [list $gid_surface($si) $gid_swap_sense($gid_sense($vi))]
            }
            set gid_volume($name) [GiD_Geometry create volume append $layer [llength $boundary_surfaces] {*}$boundary_surfaces]
        } elseif { $key == "GSUR" } {
            set fail_surface 0
            lassign $line key name sense type_or_shape_or_nurbs
            set name [Calculix::FixAutomaticName $name]
            set type_or_shape_or_nurbs [Calculix::FixAutomaticName $type_or_shape_or_nurbs]
            set operation [string tolower $sense]
            if { $operation == "add" } {
                #this really is a continuation of the previous definition whith more curves
                set more_versos_and_lines [lrange $line 3 end]
                set versos_and_lines [concat $gid_surface_versos_and_lines($name) $more_versos_and_lines]
                set sense $gid_surface_sense($name)
                set type_or_shape_or_nurbs $gid_surface_type_or_shape_or_nurbs($name)
            } else {
                set versos_and_lines [lrange $line 4 end]
            }
            set geometrical_data ""
            set gid_surface_type ""
            set type [string tolower $type_or_shape_or_nurbs]
            if { $type == "blend" } {
                set gid_surface_type coonsurface
            } else {
                #could be an entity shape (plane) or nurbs to define inner shape
                if { [info exists gid_shape_plane($type_or_shape_or_nurbs)] } {
                    set gid_surface_type plsurface
                    #set geometrical_data $gid_shape_plane($type_or_shape_or_nurbs)
                } elseif { [info exists gid_nurbssurface_data($type_or_shape_or_nurbs)] } {
                    set gid_surface_type nurbssurface
                    set geometrical_data $gid_nurbssurface_data($type_or_shape_or_nurbs)
                } else {
                    set gid_surface_type ""
                }
            }
            if { $gid_surface_type != "" } {
                set boundary_lines [list]
                foreach {vi li} $versos_and_lines {
                    set li [Calculix::FixAutomaticName $li]
                    if { [info exists gid_line($li)] } {
                        lappend boundary_lines [list $gid_line($li) $gid_sense($vi)]
                    } else {
                        if { [info exists gid_polyline($li)] } {
                            #replace the polyline by its sublines and versos
                            set polyline_lines_and_versos $gid_polyline($li)
                            if { $vi == "-" } {
                                #must reorder lines and versos
                                #must reverse lines and swap versos
                                set i 0
                                foreach item $polyline_lines_and_versos {
                                    lset polyline_lines_and_versos [list $i 1] $gid_swap_sense([lindex $item 1])
                                    incr i
                                }
                                set polyline_lines_and_versos [lreverse $polyline_lines_and_versos]
                            }
                            lappend boundary_lines {*}$polyline_lines_and_versos
                        } else {
                            set fail_surface 1
                            W "line $li doesn't exists. $line"
                            break
                        }
                    }
                }
                if { $sense == "-" } {
                    #must reverse lines and swap versos
                    set i 0
                    foreach item $boundary_lines {
                        lset boundary_lines [list $i 1] $gid_swap_sense([lindex $item 1])
                        incr i
                    }
                    set boundary_lines [lreverse $boundary_lines]
                } elseif { $sense == "+" } {
                } else {
                    W "Unexpected sense $sense. $line"
                }
                if { $fail_surface } {
                    set gid_surface($name) 0
                } else {
                    if { [info exists gid_surface($name)] && $gid_surface($name) != 0 } {
                        #the surface was created, but add will create it again with more lines (e.g. adding a hole), then delete it
                        GiD_Geometry -v2 delete surface [list $gid_surface($name)]
                        unset gid_surface($name)
                    }
                    if { $gid_surface_type == "nurbssurface" } {
                        set gid_surface($name) [GiD_Geometry create surface append nurbssurface $layer [llength $boundary_lines] \
                            {*}[lrange $geometrical_data 0 5] {*}$boundary_lines {*}[lrange $geometrical_data 6 end]]
                    } else {
                        if { $gid_surface_type == "coonsurface" } {
                            if { [llength $boundary_lines] == 4 } {
                                set gid_surface($name) [GiD_Geometry create surface append coonsurface $layer [llength $boundary_lines] {*}$boundary_lines]
                            } else {
                                #try to create a NURBS
                                set gid_surface($name) [Calculix::CreateNurbsSurfaceByBoundary $layer $boundary_lines]
                            }
                        } elseif { $gid_surface_type == "plsurface" } {
                            set gid_surface($name) [GiD_Geometry create surface append plsurface $layer [llength $boundary_lines] {*}$boundary_lines]
                        } else {
                            W "Unexpected surface type $gid_surface_type. $line"
                        }
                    }
                }
                set gid_surface_versos_and_lines($name) $versos_and_lines ;#copy it because ADD could define more lines!!
                set gid_surface_sense($name) $sense ;#copy it because ADD could define more lines!!
                set gid_surface_type_or_shape_or_nurbs($name) $type_or_shape_or_nurbs ;#copy it because ADD could define more lines!!
            } else {
                W "Unexpected surface type. $line"
            }
        } elseif { $key == "LCMB" } {
            set name [Calculix::FixAutomaticName [lindex $line 1]]
            set operation [string tolower [lindex $line 2]]
            if { $operation == "add" } {
                set versos_lines [lrange $line 3 end]
            } else {
                set versos_lines [lrange $line 2 end]
            }
            if { $operation == "add" } {
                #add more lines to the previous polyline
                set gid_lines_and_versos $gid_polyline($name)
            } else {
                set gid_lines_and_versos [list]
            }
            foreach {vi li} $versos_lines {
                set li [Calculix::FixAutomaticName $li]
                if { [info exists gid_line($li)] } {
                    lappend gid_lines_and_versos [list $gid_line($li) $gid_sense($vi)]
                } else {
                    if { [info exists gid_polyline($li)] } {
                        #replace the polyline by its sublines and versos
                        set polyline_lines_and_versos $gid_polyline($li)
                        if { $vi == "-" } {
                            #must reorder lines and versos
                            #must reverse lines and swap versos
                            set i 0
                            foreach item $polyline_lines_and_versos {
                                lset polyline_lines_and_versos [list $i 1] $gid_swap_sense([lindex $item 1])
                                incr i
                            }
                            set polyline_lines_and_versos [lreverse $polyline_lines_and_versos]
                        }
                        lappend gid_lines_and_versos {*}$polyline_lines_and_versos
                    } else {
                        set fail_polyline 1
                        W "line $li doesn't exists. $line"
                        break
                    }
                }
            }
            if { [info exists gid_point($p1)] && [info exists gid_point($p2)] } {
                #set gid_line($name) [Calculix::CreatePolyline $layer $gid_point($p1) $gid_point($p2) $gid_lines_and_versos]
                #not create a polyline because it will replace the isolated lines, and in Calculix they must continue existing!!
                #and calculix is using part of polylines for coons, like the deprecated style of GiD, to force having 4 lines
                set gid_polyline($name) $gid_lines_and_versos
            } else {
                W "point $p1 or $p2 doesn't exists"
            }
        } elseif { $key == "LINE" } {
            if { [llength $line] == 5 } {
                lassign $line key name p1 p2 div
                set name [Calculix::FixAutomaticName $name]
                set p1 [Calculix::FixAutomaticName $p1]
                set p2 [Calculix::FixAutomaticName $p2]
                if { [info exists gid_point($p1)] && [info exists gid_point($p2)] } {
                    set gid_line($name) [GiD_Geometry create line append stline $layer $gid_point($p1) $gid_point($p2)]
                } else {
                    W "point $p1 or $p2 doesn't exists"
                }
            } elseif { [llength $line] == 6 } {
                lassign $line key name p1 p2 cp_or_seq div
                set name [Calculix::FixAutomaticName $name]
                set p1 [Calculix::FixAutomaticName $p1]
                set p2 [Calculix::FixAutomaticName $p2]
                set cp_or_seq [Calculix::FixAutomaticName $cp_or_seq]
                if { [info exists gid_sequence($cp_or_seq)] } {
                    #points sequence
                    set interpolation_coordinates [list]
                    foreach point_id $gid_sequence($cp_or_seq) {
                        set point_id [Calculix::FixAutomaticName $point_id]
                        lappend interpolation_coordinates [lrange [GiD_Geometry get point $point_id] 1 3]
                    }
                    if { [info exists gid_point($p1)] && [info exists gid_point($p2)] } {
                        set gid_line($name) [GiD_Geometry create line append nurbsline $layer $gid_point($p1) $gid_point($p2) \
                            -interpolate [llength $interpolation_coordinates] {*}$interpolation_coordinates]
                    } else {
                        W "point $p1 or $p2 doesn't exists"
                    }
                } else {
                    #center point of arc
                    if { [info exists gid_point($p1)] && [info exists gid_point($p2)] } {
                        set arc_data [Calculix::GetArcData $gid_point($p1) $gid_point($p2) $gid_point($cp_or_seq)]
                        set gid_line($name) [GiD_Geometry create line append arcline $layer $gid_point($p1) $gid_point($p2) {*}$arc_data]
                    } else {
                        W "point $p1 or $p2 doesn't exists"
                    }
                }
            } else {
                W "Unexpected $line"
            }
            lappend lines_num_structured_divisions($div) $gid_line($name)
        } elseif { $key == "MESH" } {
            #e.g. MESH all
        } elseif { $key == "NURL" } {
            set name [Calculix::FixAutomaticName [lindex $line 1]]
            set operation [lindex $line 2]
            if { $operation == "DEFINE" } {
                incr count(NURL)
                lassign [lrange $line 3 7] p1 p2 degree num_points num_knots
                set p1 [Calculix::FixAutomaticName $p1]
                set p2 [Calculix::FixAutomaticName $p2]
                set gid_nurbsline($name) [list $p1 $p2 $degree $num_points $num_knots]
            } elseif { $operation == "CONTROL" } {
                #set i [lindex $line 3] ;ignore it, assume provided ordered
                set id [Calculix::FixAutomaticName [lindex $line 4]]
                if { [info exists gid_point($id)] } {
                    set xyz [lrange [GiD_Geometry get point $gid_point($id)] 1 3]
                } else {
                    set xyz [lrange $line 4 6]
                }
                lappend gid_nurbsline($name,control_points) $xyz
            } elseif { $operation == "KNOT" } {
                #set i [lindex $line 3] ;ignore it, assume provided ordered
                lappend gid_nurbsline($name,knots) [lindex $line 4]
            } elseif { $operation == "END" } {
                #build it
                lassign $gid_nurbsline($name) p1 p2 degree num_points num_knots
                if { [info exists gid_point($p1)] && [info exists gid_point($p2)] } {
                    set is_rational 0
                    set gid_line($name) [GiD_Geometry create line append nurbsline $layer $gid_point($p1) $gid_point($p2) \
                        $degree $num_points $is_rational {*}$gid_nurbsline($name,control_points) \
                        {*}[Calculix::NormalizeKnots $gid_nurbsline($name,knots)]]
                } else {
                    W "point $p1 or $p2 doesn't exists"
                }
            } else {
                W "Unexpected operation $operation. $line"
            }
        } elseif { $key == "NURS" } {
            set name [Calculix::FixAutomaticName [lindex $line 1]]
            set operation [lindex $line 2]
            if { $operation == "DEFINE" } {
                incr count(NURS)
                lassign [lrange $line 3 8] degree_u degree_v num_points_u num_points_v num_knots_u num_knots_v
                set gid_nurbssurface($name) [list $degree_u $degree_v $num_points_u $num_points_v $num_knots_u $num_knots_v]
                set m [::struct::matrix]
                $m add columns $num_points_u
                $m add rows $num_points_v
                set gid_nurbssurface($name,control_points) $m
            } elseif { $operation == "CONTROL" } {
                set column [lindex $line 3]
                set row [lindex $line 4]
                incr column -1
                incr row -1
                #must use i,j because the control points are provided unordered!!
                set id [Calculix::FixAutomaticName [lindex $line 5]]
                if { [info exists gid_point($id)] } {
                    set xyz [lrange [GiD_Geometry get point $gid_point($id)] 1 3]
                } else {
                    set xyz [lrange $line 4 6]
                }
                $gid_nurbssurface($name,control_points) set cell $column $row $xyz
            } elseif { $operation == "KNOT" } {
                #set i [lindex $line 4] ;ignore it, assume provided ordered
                set direction [string tolower [lindex $line 3]]
                if { $direction == "u" || $direction == "v" } {
                    lappend gid_nurbssurface($name,knots_$direction) [lindex $line 5]
                } else {
                    W "Unexpected direction $direction. $line"
                }
            } elseif { $operation == "END" } {
                #build it
                lassign $gid_nurbssurface($name) degree_u degree_v num_points_u num_points_v num_knots_u num_knots_v
                set is_trimmed 1
                set is_rational 0
                set gid_nurbssurface_data($name) [list $degree_u $degree_v $num_points_u $num_points_v $is_trimmed $is_rational \
                    {*}[join [lindex [$gid_nurbssurface($name,control_points) serialize] 2]] \
                    {*}[Calculix::NormalizeKnots $gid_nurbssurface($name,knots_u)] \
                    {*}[Calculix::NormalizeKnots $gid_nurbssurface($name,knots_v)]]
            } else {
                W "Unexpected operation $operation. $line"
            }
        } elseif { $key == "PNT" } {
            if { [llength $line] == 5 } {
                lassign $line key name x y z
                set name [Calculix::FixAutomaticName $name]
                set gid_point($name) [GiD_Geometry create point append $layer $x $y $z]
            } else {
                W "Unexpected $line"
            }
        } elseif { $key == "PROC" } {
            #undocummented
        } elseif { $key == "RNAM" } {
            set name [Calculix::FixAutomaticName [lindex $line 1]]
            set new_name [Calculix::FixAutomaticName [lindex $line 2]]
            if { [GiD_Groups exists $name] } {
                GiD_Groups edit rename $name $new_name
            } else {
                W "Error, group $name doesn't exists. $line"
            }
        } elseif { $key == "SEND" } {
            #to write groups to disk, ignore it
        } elseif { $key == "SEQA" } {
            set name [Calculix::FixAutomaticName [lindex $line 1]]
            set insert [string tolower [lindex $line 2]]
            if { $insert == "end" } {
                set over [string tolower [lindex $line 3]]
                if { $over == "pnt" || $over == "nod" } {
                    set names [lrange $line 4 end]
                } else {
                    set over $gid_sequence_over($name) ;#assumed set before
                    set names [lrange $line 3 end]
                }
            } elseif { $insert == "afte" || $insert == "befo" } {
                set over [string tolower [lindex $line 3]]
                if { $over == "pnt" || $over == "nod" } {
                    set insert_name [lindex $line 4]
                    set names [lrange $line 5 end]
                } else {
                    set over $gid_sequence_over($name) ;#assumed set before
                    set insert_name [lindex $line 3]
                    set names [lrange $line 4 end]
                }
            } else {
                set insert ""
                set over [string tolower [lindex $line 2]]
                set names [lrange $line 3 end]
            }
            if { $over == "pnt" || $over == "nod" } {
                set over_letter [string index $over 0] ;#get only the first letter
                set gid_array_variable $gid_array_over($over_letter)
                set selection [list]
                foreach id $names {
                    set id [Calculix::FixAutomaticName $id]
                    lappend selection [set ${gid_array_variable}($id)]
                }
                if { $insert == "end" } {
                    set selection [linsert $gid_sequence($name) end {*}$selection]
                } elseif { $insert == "afte" || $insert == "befo" } {
                    set insert_gid_id [set ${gid_array_variable}($insert_name)]
                    set index [lsearch $gid_sequence($name) $insert_gid_id]
                    if { $index == -1 } {
                        W "Error, insertion item $insert_name not found. $line"
                    }
                    if { $insert == "afte" } {
                        incr index
                    }
                    set selection [linsert $gid_sequence($name) $index {*}$selection]
                }
                set gid_sequence($name) $selection
                set gid_sequence_over($name) $over
            } else {
                W "Unexpected $over. $line"
            }
        } elseif { $key == "SETA" } {
            set name [Calculix::FixAutomaticName [lindex $line 1]]
            set over [string tolower [lindex $line 2]]
            set over_letter [string index $over 0]
            if { ![GiD_Groups exists $name] } {
                GiD_Groups create $name
            }
            if { $over == "all" } {
                set all_geometry [list [GiD_Geometry list point 1:end] [GiD_Geometry list line] \
                    [GiD_Geometry list surface] [GiD_Geometry list volume]]
                GiD_EntitiesGroups assign $name all_geometry $all_geometry
            } else {
                if { $over_letter == "c" } {
                    #polylines
                    set over_letter l
                    set selection [list]
                    foreach id [lrange $line 3 end] {
                        foreach item $gid_polyline($id) {
                            lappend selection [lindex $item 0]
                        }
                    }
                    GiD_EntitiesGroups assign $name $gid_entity_over($over) $selection
                } elseif { [info exists gid_array_over($over_letter)] } {
                    set gid_array_variable $gid_array_over($over_letter)
                    set selection [list]
                    foreach id [lrange $line 3 end] {
                        set id [Calculix::FixAutomaticName $id]
                        lappend selection [set ${gid_array_variable}($id)]
                    }
                    GiD_EntitiesGroups assign $name $gid_entity_over($over_letter) $selection
                } else {
                    W "unexpected over $over. $line"
                }
            }
        } elseif { $key == "SETC" } {
            #close a set
            set name [Calculix::FixAutomaticName [lindex $line 1]]
            set gid_last_opened_set($name) 0
        } elseif { $key == "SETO" } {
            #open a set
            set name [Calculix::FixAutomaticName [lindex $line 1]]
            set gid_last_opened_set($name) 1
        } elseif { $key == "SETR" } {
            #remove entities
            set name [Calculix::FixAutomaticName [lindex $line 1]]
            set over [string tolower [lindex $line 2]]
            set over_letter [string index $over 0]
            if { [GiD_Groups exists $name] } {
                if { $over == "all" } {
                    GiD_EntitiesGroups unassign $name all_geometry
                } elseif { [info exists gid_array_over($over_letter)] } {
                    set gid_array_variable $gid_array_over($over_letter)
                    set selection [list]
                    foreach id [lrange $line 3 end] {
                        set id [Calculix::FixAutomaticName $id]
                        lappend selection [set ${gid_array_variable}($id)]
                    }
                    if { [llength $selection] } {
                        GiD_EntitiesGroups unassign $name $gid_entity_over($over_letter) $selection
                    }
                } else {
                    W "unexpected over $over. $line"
                }
            } else {
                W "Error, group $name doesn't exists. $line"
            }
        } elseif { $key == "TEXT" } {
            #set text [lindex $line 1]
        } elseif { $key == "SHPE" } {
            set name [Calculix::FixAutomaticName [lindex $line 1]]
            set type [lindex $line 2]
            if { $type == "PLN" } {
                set plane_points [list]
                foreach id [lrange $line 3 5] {
                    set id [Calculix::FixAutomaticName $id]
                    lappend plane_points $gid_point($id)
                }
                #store the 3 point ids
                set gid_shape_plane($name) $plane_points
            } else {
                W "unexpecte shape type. $line"
            }
        } elseif { $key == "PLOT" || $key == "PLUS" } {
            #ignore it
        } else {
            #W "key $key not implemented. $line"
            incr gid_unimplemented($key)
        }
    }
    close $fp
    set lines_divisions [lsort -integer [array names lines_num_structured_divisions]]
    if { [llength $lines_divisions] } {
        if { [lindex $lines_divisions 0] > 0 && [lindex $lines_divisions end] < 1000 } {
            #it seems and error (example in airfoil/rae2822.fbd) has negative and huge divisions!!
            foreach num_divisions $lines_divisions {
                set selection [lsort -integer -unique $lines_num_structured_divisions($num_divisions)]
                GiD_Process Mescape Meshing Structured Lines $num_divisions {*}$selection escape escape
            }
        }
    }
    if { 0 } {
        #delete isolated points
        set id0 [expr {$last_point_start+1}]
        GiD_Process Mescape Geometry Delete Points $id0:end escape escape
    }
    if { [GiD_Set AutoCollapseAfterImport] } {
        #do a collapse and repair
        GiD_Process Mescape Utilities Collapse Model Yes escape escape
        GiD_Process Mescape Utilities Repair Yes escape escape
    } else {
        if { $count(NURL) } {
            #points are defined repeated for each NURL curve without share them (example cad/halter.fbd) and then surfaces don't close!!
            set id0 [expr {$last_point_start+1}]
            GiD_Process Mescape Utilities Collapse Points $id0:end escape escape
        }
        if { $count(NURS) } {
            #orientation of curves is wrong for surfaces GSUR created from nurbs surfaces NURS(example cad/halter.fbd)!!
            GiD_Process Mescape Utilities Repair Yes escape escape
        }
    }
    set unimplemented_keys [lsort -dictionary [array names gid_unimplemented]]
    if { [llength $unimplemented_keys] }  {
        W "Unimplemented keys:$unimplemented_keys"
    }
    return $fail
}

proc Calculix::ReadPre0 { filenames } {
    variable percent_done
    variable stop
    set fail 0
    if { [llength $filenames] == 1 && [file isdirectory $filenames] } {
        #now the window only select files, not directories, but it is possible to invoke with folder
        #from Tcl command
        set filenames [Calculix::GetFilesRecursive $filenames]
    }
    set ::Calculix::stop 0
    set ::Calculix::percent_done 0
    set percent_max [expr {3*[llength $filenames]}]
    ::GidUtils::CreateAdvanceBar [_ "Loading file"] [_ "Process"] ::Calculix::percent_done $percent_max {set ::Calculix::stop 1}
    foreach filename $filenames {
        if { $::Calculix::stop } {
            break
        }
        if { [file exists $filename] } {
            set res ""
            if { [catch { set local_fail [Calculix::ReadPreGeometry $filename] } err options] } {
                #set fail 1
                #W $err
                GidUtils::DeleteAdvanceBar
                return -options $options $err
            } else {
                if { $local_fail } {
                    set fail 1
                }
                GiD_RaiseEvent GiD_Event_AfterOpenFile $filename CALCULIX_FBD_FORMAT $local_fail
            }

        }
        incr ::Calculix::percent_done
    }
    GidUtils::DeleteAdvanceBar
    return $fail
}

#read and finally redraw showing what is read
proc Calculix::ReadPre { filenames } {
    GidUtils::WaitState .gid
    GidUtils::DisableGraphics
    if { [catch { set fail [Calculix::ReadPre0 $filenames] } err options] } {
        set fail 1
    }
    GidUtils::EnableGraphics
    GidUtils::EndWaitState .gid
    if { $fail } {
        return -options $options $err
    } else {
        GiD_Process 'Zoom Frame escape
        GidUtils::UpdateWindow GROUPS
        GidUtils::SetWarnLine [_ "Geometry read"]
    }
    return
}

#to drag and drop a file
proc Calculix::ReadPreSingleFile { filename } {
    GidUtils::EvalAndEnterInBatch Calculix::ReadPre [list $filename]
}

proc Calculix::ReadPreW { } {
    variable _merge
    set multiple 1
    set filenames [MessageBoxGetFilename file read [_ "Read Calculix geometry"] \
        {} [list [list {Calculix} {.fbd }] [list [_ "All files"] {.*}]] {} $multiple \
        [list [_ "Import options"] Calculix::MoreImportOptions]]
    if {$filenames == ""} {
        return
    }
    if { !$_merge } {
        GiD_Process Mescape Files New No escape ;#to avoid previous geometry
    }
    GidUtils::EvalAndEnterInBatch Calculix::ReadPre $filenames
}

#########################################################################################
######################################### READ POST #####################################
#########################################################################################

proc Calculix::InitReadPost { } {

    variable GidPostElementType
    variable GidPostElementNNode
    variable GidPostGaussPoint1

    set GidPostElementType(11) Line
    set GidPostElementType(7) Triangle
    set GidPostElementType(9) Quadrilateral
    set GidPostElementType(3) Tetrahedra
    set GidPostElementType(1) Hexahedra
    set GidPostElementType(2) Prism
    #quadratic
    set GidPostElementType(12) Line
    set GidPostElementType(8) Triangle
    set GidPostElementType(10) Quadrilateral
    set GidPostElementType(6) Tetrahedra
    set GidPostElementType(4) Hexahedra
    set GidPostElementType(5) Prism

    set GidPostElementNNode(11) 2
    set GidPostElementNNode(7) 3
    set GidPostElementNNode(9) 4
    set GidPostElementNNode(3) 4
    set GidPostElementNNode(1) 8
    set GidPostElementNNode(2) 6
    #quadratic
    set GidPostElementNNode(12) 3
    set GidPostElementNNode(8) 6
    set GidPostElementNNode(10) 8
    set GidPostElementNNode(6) 10
    set GidPostElementNNode(4) 20
    set GidPostElementNNode(5) 15

    set GidPostGaussPoint1(11) GP_LINEAR_1
    set GidPostGaussPoint1(7) GP_TRIANGLE_1
    set GidPostGaussPoint1(9) GP_QUADRILATERAL_1
    set GidPostGaussPoint1(3) GP_TETRAHEDRA_1
    set GidPostGaussPoint1(1) GP_HEXAHEDRA_1
    set GidPostGaussPoint1(2) GP_PRISM_1
    #quadratic
    set GidPostGaussPoint1(12) GP_LINEAR_1
    set GidPostGaussPoint1(8) GP_TRIANGLE_1
    set GidPostGaussPoint1(10) GP_QUADRILATERAL_1
    set GidPostGaussPoint1(6) GP_TETRAHEDRA_1
    set GidPostGaussPoint1(4) GP_HEXAHEDRA_1
    set GidPostGaussPoint1(5) GP_PRISM_1
}

proc Calculix::GetGidPostElementType { element_code } {
    variable GidPostElementType
    if { [info exists GidPostElementType($element_code)] } {
        return $GidPostElementType($element_code)
    } else {
        WarnWinText [_ "GetGidPostElementType. Unexpected element_code '%s'" $element_code]
        return ""
    }
}

proc Calculix::GetGidPostElementNNode { element_code } {
    variable GidPostElementNNode
    if { [info exists GidPostElementNNode($element_code)] } {
        return $GidPostElementNNode($element_code)
    } else {
        WarnWinText [_ "GetGidPostElementNNode. Unexpected element_code '%s'" $element_code]
        return 0
    }
}

proc Calculix::GetGidPostGaussPoint1Type { element_code } {
    variable GidPostGaussPoint1
    if { [info exists GidPostGaussPoint1($element_code)] } {
        return $GidPostGaussPoint1($element_code)
    } else {
        WarnWinText [_ "GetGidPostGaussPoint1Type. Unexpected element_code '%s'" $element_code]
        return ""
    }
}


proc Calculix::IncrListItems { list increment } {
    set xs {}
    foreach item $list {
        lappend xs [incr item $increment]
    }
    return $xs
}

proc Calculix::CreateGiD_MeshPost { meshname gidpost_element_type gidpost_element_nnode gidpost_node_ids \
    gidpost_nodes gidpost_element_ids gidpost_elements offset_nodes offset_elements } {

    if { $offset_nodes } {
        set gidpost_node_ids [Calculix::IncrListItems $gidpost_node_ids $offset_nodes]
        set gidpost_elements [Calculix::IncrListItems $gidpost_elements $offset_nodes]
    }
    if { $offset_elements } {
        set gidpost_element_ids [Calculix::IncrListItems $gidpost_element_ids $offset_elements]
    }
    GiD_MeshPost create $meshname $gidpost_element_type $gidpost_element_nnode $gidpost_node_ids $gidpost_nodes \
        $gidpost_element_ids $gidpost_elements

}

proc Calculix::ReadPostMeshAndResults { filename } {
    variable percent_done

    #some files seems wrong: negative floats with 3 exponentd digits instead of 12 span 13 characters
    set exponent_lenght 0 ;#0 unknown, to test only until is determined to 2 or 3

    if { ![file exists $filename] } {
        WarnWin [_ "File %s not found" $filename]
        return [list 0 0]
    }

    set fp [open $filename r]
    if { $fp == "" } {
        WarnWin [_ "Cannot open file '%s'" $filename]
        return [list 0 0]
    }

    set num_meshes 0
    set num_results 0
    set offset_nodes [GiD_Info Mesh -post MaxNumNodes]
    set offset_elements [GiD_Info Mesh -post MaxNumElements]

    set gidpost_node_ids ""
    set gidpost_nodes ""

    set mode 0;#0 == short format, 1==long format, 2==binary format
    while { ![eof $fp] } {
        gets $fp line
        set key [string trim [string range $line 1 4]]
        set code [string index $line 5]
        if { $key == "1" } {
            if { $code == "C" } {
                #Model Header Record
                #Format:(1X,'   1','C',A6)
                set model_name [string trim [string range $line 6 11]]
            } elseif { $code == "U" } {
                #User Header Record
                #Format:(1X,'    1','U',A66)
                lappend job_info [string trim [string range $line 6 71]]
            } elseif { $code == "P" } {
                #Parameter Header Record
                #is a pair keyword value
                set string [string trim [string range $line 6 71]]
            } else {
                set to_set_breakpoint 1
            }
        } elseif { $key == "2" } {
            if { $code == "C" } {
                # Nodal Point Coordinate Block
                #Format:(1X,'   2','C',18X,I12,37X,I1)
                set num_nodes_block [string trim [string range $line 24 35]] ;#sometimes missing
                set mode [string index $line 73]
                if { $mode == "" } {
                    set mode 0
                }
                set key -1
                set i 0
                if { $mode == 0 } {
                    #Short Format:(1X,'-1',I5,3E12.5)
                    #Long Format:(1X,'-1',I10,3E12.5)
                    set id_size 5
                } elseif { $mode == 1 } {
                    set id_size 10
                } else {
                    #binary format not implemented yet
                    break
                }
                while { $key != "-3" } {
                    gets $fp line
                    set key [string range $line 1 2]
                    if { $key == "-1" } {
                        set item_begin 3
                        set item_end [expr {$item_begin+$id_size}]
                        set node_id [string trim [string range $line $item_begin $item_end-1]]
                        set item_begin $item_end
                        set item_end [expr {$item_begin+12}]
                        if { $exponent_lenght == 0 } {
                            #test only once
                            set txt [string range $line $item_begin end]
                            set last_E [string last E $txt]
                            if { $last_E != -1 } {
                                set distance [expr {[string length $txt]-$last_E-2}]
                                if { $distance == 2 || $distance == 3 } {
                                    set exponent_lenght $distance
                                } else {
                                    set exponent_lenght 0 ;#continue in doubth
                                }
                            } else {
                                set exponent_lenght 0 ;#continue in doubth
                            }
                        }

                        if { $exponent_lenght == 3 && [string index $line $item_begin] == "-" } {
                            #negative floats span 13 digits instead 12!!
                            incr item_end
                        }
                        set node_x [string trim [string range $line $item_begin $item_end-1]]
                        set item_begin $item_end
                        set item_end [expr {$item_begin+12}]
                        if { $exponent_lenght == 3 && [string index $line $item_begin] == "-" } {
                            #negative floats span 13 digits instead 12!!
                            incr item_end
                        }
                        set node_y [string trim [string range $line $item_begin $item_end-1]]
                        set item_begin $item_end
                        set item_end [expr {$item_begin+12}]
                        if { $exponent_lenght == 3 && [string index $line $item_begin] == "-" } {
                            #negative floats span 13 digits instead 12!!
                            incr item_end
                        }
                        set node_z [string trim [string range $line $item_begin $item_end-1]]
                        lappend gidpost_node_ids $node_id
                        lappend gidpost_nodes $node_x $node_y $node_z
                    }
                }
                incr ::Calculix::percent_done
            } elseif { $code == "U" } {
                #User Header Record
                set job_info [string range $line 6 end]
            } else {
                set to_set_breakpoint 1
            }
        } elseif { $key == "3" } {
            if { $code == "C" } {
                #Element Definition Block
                #Format:(1X,'   3','C',18X,I12,37X,I1)
                set num_elements_block [string trim [string range $line 24 35]]  ;#sometimes missing
                set mode [string index $line 73]
                if { $mode == "" } {
                    set mode 0
                }
                set key -1
                if { $mode == 0 } {
                    #Short Format:(1X,'-1',I5,3I5)
                    set id_size 5
                    set nodes_by_line 15  ;#Short Format:(1X,'-2',15I5)
                } elseif { $mode == 1 } {
                    #Long Format:(1X,'-1',I10,3I5)
                    set id_size 10
                    set nodes_by_line 10
                } else {
                    #binary format not implemented yet
                    break
                }

                gets $fp line
                set key [string range $line 1 2]
                while { $key != -3 } {
                    set item_begin 3
                    set item_end [expr {$item_begin+$id_size}]
                    if { $key == -1 } {
                        set element_id [string trim [string range $line $item_begin $item_end-1]]
                        set item_begin $item_end
                        set item_end [expr {$item_begin+5}]
                        set element_type [string trim [string range $line $item_begin $item_end-1]]
                        #GiD_MeshPost only implemented now for linear meshes, not quadratic...
                        if { $element_type == 4 } {
                            set element_type 1
                        } elseif { $element_type == 5 } {
                            set element_type 2
                        } elseif { $element_type == 6 } {
                            set element_type 3
                        } elseif { $element_type == 8 } {
                            set element_type 7
                        } elseif { $element_type == 10 } {
                            set element_type 9
                        } elseif { $element_type == 12 } {
                            set element_type 11
                        }
                        set item_begin $item_end
                        set item_end [expr {$item_begin+5}]
                        set element_group [string trim [string range $line $item_begin $item_end-1]]
                        set item_begin $item_end
                        set item_end [expr {$item_begin+5}]
                        set element_material [string trim [string range $line $item_begin $item_end-1]]
                        set gidpost_element_nnode [Calculix::GetGidPostElementNNode $element_type]
                        set element_nodes ""
                        gets $fp line
                        set key [string range $line 1 2]
                    } elseif { $key == -2 } {
                        #continuation line
                        #Long Format:(1X,'-2',10I10)
                        #Values: KEY,NODE,NODE,NODE,NODE
                    } elseif { $key == -3 } {
                        break
                    }

                    set item_end 3
                    for {set i 0} {$i < $nodes_by_line } {incr i} {
                        if { [llength $element_nodes] < $gidpost_element_nnode } {
                            set item_begin $item_end
                            set item_end [expr {$item_begin+$id_size}]
                            lappend element_nodes [string trim [string range $line $item_begin $item_end-1]]
                        } else {
                            break
                        }
                    }
                    gets $fp line
                    set key [string range $line 1 2]
                    if { $key != -2 } {
                        #-2 continuation line
                        lappend elements($element_type) {*}$element_nodes
                        lappend element_ids($element_type) $element_id
                    }
                }
                set different_element_types [lsort -unique -integer [array names elements]]
                foreach element_type $different_element_types {
                    set gidpost_element_type [Calculix::GetGidPostElementType $element_type]
                    if { $gidpost_element_type != "" } {
                        set meshname [Calculix::GetPostMeshNameAutomatic $gidpost_element_type] ;#??
                        set gidpost_element_nnode [Calculix::GetGidPostElementNNode $element_type]
                        Calculix::CreateGiD_MeshPost $meshname \
                            $gidpost_element_type $gidpost_element_nnode $gidpost_node_ids $gidpost_nodes \
                            $element_ids($element_type) $elements($element_type) $offset_nodes $offset_elements
                        incr num_meshes
                    }
                }
            } else {
                #code
                set to_set_breakpoint 1
            }
            incr ::Calculix::percent_done
        } elseif { $key == "100" } {
            if { $code == "C" } {
                set gidpost_ids ""
                array unset gidpost_values_component
                #Nodal Results Block
                #Format:(1X,'100','C',6A1,E12.5,I12,20A1,I2,I5,10A1,I2)
                #Values: KEY,CODE,SETNAME,VALUE,NUMNOD,TEXT,analysis_type,NUMSTP,ANALYS,FORMAT
                set item_end 6
                set item_begin $item_end
                set item_end [expr {$item_begin+6}]
                set result_name [string trim [string range $line $item_begin $item_end-1]]
                set item_begin $item_end
                set item_end [expr {$item_begin+12}]
                if { $exponent_lenght == 3 && [string index $line $item_begin] == "-" } {
                    #negative floats span 13 digits instead 12!!
                    incr item_end
                }
                set value [string trim [string range $line $item_begin $item_end-1]]
                set item_begin $item_end
                set item_end [expr {$item_begin+12}]
                set numnod [string trim [string range $line $item_begin $item_end-1]]
                set item_begin $item_end
                set item_end [expr {$item_begin+20}]
                set text [string trim [string range $line $item_begin $item_end-1]]
                set item_begin $item_end
                set item_end [expr {$item_begin+2}]
                set analysis_type [string trim [string range $line $item_begin $item_end-1]]
                #analysis_type = Analysis type
                #0 static
                #1 time step
                #2 frequency
                #3 load step
                #4 user named
                set item_begin $item_end
                set item_end [expr {$item_begin+5}]
                set numstp [string trim [string range $line $item_begin $item_end-1]]
                set item_begin $item_end
                set item_end [expr {$item_begin+10}]
                set analys [string trim [string range $line $item_begin $item_end-1]]
                set item_begin $item_end
                set item_end [expr {$item_begin+2}]
                set mode [string trim [string range $line $item_begin $item_end-1]]
                if { $mode == "" } {
                    set mode 0
                }

                #Format:(1X,I2,2X,8A1,2I5)
                #Values: KEY, NAME, NCOMPS, IRTYPE
                gets $fp line
                set key [string range $line 1 2] ;#-4
                set result_name [string trim [string range $line 3 12]]
                set ncomps [string trim [string range $line 13 17]]
                set irtype [string trim [string range $line 18 22]]
                #IRTYPE = 1 Nodal data, material independent
                #2 Nodal data, material dependant
                #3 Element data at nodes (not used)

                array unset component_name
                array unset menu
                array unset ictype
                array unset icind1
                array unset icind2
                array unset iexist
                array unset icname
                #Format:(1X,I2,2X,8A1,5I5,8A1)
                #Values: KEY, NAME, MENU, ICTYPE, ICIND1, ICIND2, IEXIST, ICNAME
                for {set icomp 0} {$icomp<$ncomps} {incr icomp} {
                    gets $fp line
                    set key [string range $line 1 2] ;#-5
                    set component_name($icomp) [string trim [string range $line 5 12]]
                    set menu($icomp) [string trim [string range $line 13 17]]
                    set ictype($icomp) [string trim [string range $line 18 22]]
                    set icind1($icomp) [string trim [string range $line 23 27]]
                    set icind2($icomp) [string trim [string range $line 28 32]]
                    set iexist($icomp) [string trim [string range $line 33 37]]
                    set icname($icomp) [string trim [string range $line 38 45]]
                }
                if { $ncomps == 4 && $ictype(3) == 2 && $iexist(3) == 1} {
                    #iexist == 1: data are to be calculated by predefined functions (not used)
                    #to read 3 items instead of 4, the last one is the vector modulus, self-calculated
                    set ncomps 3
                }
                #Where: KEY = -5
                #NAME = Entity name to be used in the menu for this comp.
                #MENU = 1
                #ICTYPE = Type of entity
                #1 scalar
                #2 vector with 3 components
                #4 matrix
                #12 vector with 3 amplitudes and 3 phase-angles in degree
                #ICIND1 = sub-component index or row number
                #ICIND2 = column number for ICTYPE=4
                #IEXIST = 0 data are provided
                #1 data are to be calculated by predefined functions (not used)
                #2 as 0 but earmarked
                #ICNAME = Name of the predefined calculation (not used)
                #ALL calculate the total displacement if ICTYPE=2

                gets $fp line
                set key [string range $line 1 2]

                while { $key == -6 } {
                    #optional data -6
                    #Format:(1X,I2,2I5,20I3)
                    #Values: KEY,IRECTY,NUMCPS,(LSTCPS(I),I=1,NUMCPS)
                    #Where: KEY = -6
                    #IRECTY = Record variant identification number
                    #NUMCPS = Number of components
                    #LSTCPS = For each variant component, the position of the
                    #corresponding component in the attribute definition
                    gets $fp line
                    set key [string range $line 1 2]
                }

                if { $mode == 0 } {
                    #Short Format:(1X,I2,I5,6E12.5)
                    set id_size 5
                } elseif { $mode == 1 } {
                    #Long Format:(1X,I2,I10,6E12.5)
                    set id_size 10
                } else {
                    #binary format not implemented yet
                    break
                }

                while { $key != -3 } {
                    set item_begin 3
                    set item_end [expr {$item_begin+$id_size}]
                    if { $key == -1 } {
                        set node_id [string trim [string range $line $item_begin $item_end-1]]
                        set values ""
                    } elseif { $key == -2 } {
                        #continuation line
                    } elseif { $key == -3 } {
                        break
                    }
                    for {set i 0} {$i < 6 } {incr i} {
                        if { [llength $values] < $ncomps } {
                            set item_begin $item_end
                            set item_end [expr {$item_begin+12}]
                            if { $exponent_lenght == 3 && [string index $line $item_begin] == "-" } {
                                #negative floats span 13 digits instead 12!!
                                incr item_end
                            }
                            lappend values [string trim [string range $line $item_begin $item_end-1]]
                        } else {
                            break
                        }
                    }
                    gets $fp line
                    set key [string range $line 1 2]
                    if { $key != -2 } {
                        #-2 continuation line
                        lappend gidpost_ids $node_id
                        set i 0
                        foreach item $values {
                            lappend gidpost_values_component($i) $item
                            incr i
                        }
                    }
                }

                if { $ncomps == 0 } {
                    #do nothing
                } else {
                    set gidpost_resultname $result_name
                    if { $analysis_type == 0 } {
                        set gidpost_analysisname static
                    } elseif { $analysis_type == 1 } {
                        set gidpost_analysisname time_step
                    } elseif { $analysis_type == 2 } {
                        set gidpost_analysisname frequency
                    } elseif { $analysis_type == 3 } {
                        set gidpost_analysisname load_step
                    } elseif { $analysis_type == 4 } {
                        set gidpost_analysisname user_named
                    } else {
                        set gidpost_analysisname unknown
                    }
                    set gidpost_step $value
                    if { $ictype(0) == 1 } {
                        set gidpost_type Scalar
                    } elseif { $ictype(0) == 2 } {
                        set gidpost_type Vector
                        set gidpost_component_names [list $component_name(0) $component_name(1) $component_name(2)]
                    } elseif { $ictype(0) == 4 } {
                        if { $ncomps == 6 } {
                            set gidpost_type Matrix
                            set gidpost_component_names [list $component_name(0) $component_name(1) $component_name(2) \
                                $component_name(3) $component_name(4) $component_name(5)]
                        } elseif { $ncomps == 12 } {
                            #already not exists in GiD ComplexMatrix result !!
                            set gidpost_type ComplexMatrix
                            set gidpost_component_names [list $component_name(0) $component_name(1) $component_name(2) \
                                $component_name(3) $component_name(4) $component_name(5) \
                                $component_name(6) $component_name(7) $component_name(8) \
                                $component_name(9) $component_name(10) $component_name(11)]
                        }
                    } elseif { $ictype(0) == 12 } {
                        #12 vector with 3 amplitudes and 3 phase-angles in degrees
                        set gidpost_type ComplexVector
                        set gidpost_component_names [list $component_name(0) $component_name(1) $component_name(2) \
                            $component_name(3) $component_name(4) $component_name(5)]
                        #must convert amplitudes from degrees to Imaginary part of complex vector...
                    } else {
                        set gidpost_type unknown
                        set gidpost_component_names ""
                    }
                    if { $irtype == 1 } {
                        #1 Nodal data, material independent
                        set gidpost_over OnNodes
                    } elseif { $irtype == 2 } {
                        #2 Nodal data, material dependant
                        set gidpost_over OnNodes
                    } elseif { $irtype == 3 } {
                        #3 Element data at nodes (not used)
                        set gidpost_over OnGaussPoints
                        set gidpost_gauss_point_name [Calculix::GetGidPostGaussPoint1Type $element_type]
                    } else {
                        set gidpost_over unknown
                    }


                    if { $offset_nodes } {
                        set gidpost_ids [Calculix::IncrListItems $gidpost_ids $offset_nodes]
                    }
                    if { $ncomps != [llength [array names gidpost_values_component]] } {
                        WarnWinText "Result $gidpost_resultname ncomps=$ncomps but values have llength [array names gidpost_values_component]] components"
                    }
                    if { $gidpost_type == "Scalar" } {
                        for {set i 0} {$i < $ncomps} {incr i} {
                            set gidpost_resultname_tmp $gidpost_resultname//$component_name($i)
                            set gidpost_result [list Result $gidpost_resultname_tmp $gidpost_analysisname $gidpost_step $gidpost_type $gidpost_over]
                            if { $gidpost_over == "OnGaussPoints" } {
                                lappend gidpost_result $gidpost_gauss_point_name
                            }
                            set gidpost_component_names [list $component_name($i)]
                            set gidpost_some_values [list $gidpost_ids [list $gidpost_values_component($i)]]
                            #GiD_Result create -array $gidpost_result [list ComponentNames {*}$gidpost_component_names] [list Unit $unit_name] $gidpost_some_values
                            GiD_Result create -array $gidpost_result [list ComponentNames {*}$gidpost_component_names] $gidpost_some_values
                            incr num_results
                            unset gidpost_some_values
                        }
                    } elseif { $gidpost_type == "ComplexVector" } {
                        #GiD ComplexVector result exists in GiD, but not use it because i am not sure
                        #about how to convert amplitude-phase to Re-Im
                        set gidpost_type_tmp Vector
                        set pos 0
                        foreach part [list [_ "Magnitude"] [_ "Phase"]] {
                            set gidpost_resultname_tmp $gidpost_resultname//$part
                            set gidpost_result [list Result $gidpost_resultname_tmp $gidpost_analysisname $gidpost_step $gidpost_type_tmp $gidpost_over]
                            if { $gidpost_over == "OnGaussPoints" } {
                                lappend gidpost_result $gidpost_gauss_point_name
                            }
                            set gidpost_component_names_tmp [lrange $gidpost_component_names $pos [expr {$pos+2}]]

                            set gidpost_values [list ]
                            for {set i 0} {$i < 3} {incr i} {
                                lappend gidpost_values $gidpost_values_component([expr {$pos+$i}])
                            }
                            set gidpost_some_values [list $gidpost_ids $gidpost_values]
                            unset gidpost_values
                            #GiD_Result create -array $gidpost_result [list ComponentNames {*}$gidpost_component_names_tmp] [list Unit $unit_name] $gidpost_some_values
                            GiD_Result create -array $gidpost_result [list ComponentNames {*}$gidpost_component_names_tmp] $gidpost_some_values
                            incr num_results
                            unset gidpost_some_values
                            incr pos 3
                        }
                    } elseif { $gidpost_type == "ComplexMatrix" } {
                        #already not exists in GiD ComplexMatrix result !!
                        set gidpost_type_tmp Matrix
                        set pos 0
                        foreach part [list [_ "Magnitude"] [_ "Phase"]] {
                            set gidpost_resultname_tmp $gidpost_resultname//$part
                            set gidpost_result [list Result $gidpost_resultname_tmp $gidpost_analysisname $gidpost_step $gidpost_type_tmp $gidpost_over]
                            if { $gidpost_over == "OnGaussPoints" } {
                                lappend gidpost_result $gidpost_gauss_point_name
                            }
                            set gidpost_component_names_tmp [lrange $gidpost_component_names $pos [expr {$pos+5}]]

                            set gidpost_values [list ]
                            for {set i 0} {$i < 6} {incr i} {
                                lappend gidpost_values $gidpost_values_component([expr {$pos+$i}])
                            }
                            set gidpost_some_values [list $gidpost_ids $gidpost_values]
                            unset gidpost_values
                            #GiD_Result create -array $gidpost_result [list ComponentNames {*}$gidpost_component_names_tmp] [list Unit $unit_name] $gidpost_some_values
                            GiD_Result create -array $gidpost_result [list ComponentNames {*}$gidpost_component_names_tmp] $gidpost_some_values
                            incr num_results
                            unset gidpost_some_values
                            incr pos 6
                        }
                    } else {
                        set gidpost_result [list Result $gidpost_resultname $gidpost_analysisname $gidpost_step $gidpost_type $gidpost_over]
                        if { $gidpost_over == "OnGaussPoints" } {
                            lappend gidpost_result $gidpost_gauss_point_name
                        }
                        set gidpost_values [list ]
                        for {set i 0} {$i < $ncomps} {incr i} {
                            lappend gidpost_values $gidpost_values_component($i)
                        }
                        set gidpost_some_values [list $gidpost_ids $gidpost_values]
                        unset gidpost_values
                        #GiD_Result create -array $gidpost_result [list ComponentNames {*}$gidpost_component_names] [list Unit $unit_name] $gidpost_some_values
                        GiD_Result create -array $gidpost_result [list ComponentNames {*}$gidpost_component_names] $gidpost_some_values
                        incr num_results
                        unset gidpost_some_values
                    }
                }
                unset gidpost_ids
                array unset gidpost_values_component
            } else {
                set to_set_breakpoint 1
            }
        } elseif { $key == "99" } {
            #line == 9999
            break
        } else {
            #key
            set to_set_breakpoint 1
            break
        }
    }
    close $fp
    return [list $num_meshes $num_results]
}

#to get an unused name, based on label
proc Calculix::GetPostMeshNameAutomatic { label } {
    set names [concat [GiD_Info postprocess get all_surfacesets] [GiD_Info postprocess get all_volumesets]]
    if { [lsearch $names $label] == -1 } {
        set name $label
    } else {
        set i 1
        while { 1 } {
            set name $label_$i
            if { [lsearch $names $label] == -1 } {
                break
            }
        }
    }
    return $name
}

proc Calculix::GetPostResultNameAutomatic { label } {
    set analysis [GiD_Info postprocess get cur_analysis]
    set time_step [GiD_Info postprocess get cur_step]
    set names [GiD_Info postprocess get results_list Contour_Fill $analysis $time_step]
    if { [lsearch $names $label] == -1 } {
        set name $label
    } else {
        set i 1
        while { 1 } {
            set name $label_$i
            if { [lsearch $names $label] == -1 } {
                break
            }
        }
    }
    return $name
}

proc Calculix::GetFilesRecursive { dirname } {
    set names [glob -nocomplain -directory $dirname -types f *.frd]
    foreach name [glob -nocomplain -directory $dirname -types d *] {
        lappend names {*}[Calculix::GetFilesRecursive $name]
    }
    return $names
}

proc Calculix::ReadPost0 { filenames } {
    variable percent_done
    variable stop

    lassign {0 0} num_meshes num_results

    if { [llength $filenames] == 1 && [file isdirectory $filenames] } {
        #now the window only select files, not directories, but it is possible to invoke with folder
        #from Tcl command
        set filenames [Calculix::GetFilesRecursive $filenames]
    }
    set ::Calculix::stop 0
    set ::Calculix::percent_done 0
    set percent_max [expr {3*[llength $filenames]}]
    ::GidUtils::CreateAdvanceBar [_ "Loading file"] [_ "Process"] ::Calculix::percent_done $percent_max {set ::Calculix::stop 1}
    foreach filename $filenames {
        if { $::Calculix::stop } {
            break
        }
        if { [file exists $filename] } {
            set res ""
            if { [catch { set res [Calculix::ReadPostMeshAndResults $filename] } err options] } {
                #on error must delete bar!!
                GidUtils::DeleteAdvanceBar
                return -options $options $err
            } else {
                lassign $res num_meshes_local num_results_local
                incr num_meshes $num_meshes_local
                incr num_results $num_results_local
                GiD_RaiseEvent GiD_Event_AfterOpenFile $filename CALCULIX_FRD_FORMAT 0
            }

        }
        incr ::Calculix::percent_done
    }
    GidUtils::DeleteAdvanceBar
    return [list $num_meshes $num_results]
}

#read and finally redraw showing what is read
proc Calculix::ReadPost { filenames } {
    GidUtils::WaitState .gid
    GidUtils::DisableGraphics
    if { [catch { set res [Calculix::ReadPost0 $filenames] } msg options] } {
        set num_meshes 0
        set num_results 0
        WarnWin "$msg $::errorInfo"
    } else {
        lassign $res num_meshes num_results
    }
    GidUtils::EnableGraphics
    GidUtils::EndWaitState .gid
    if { $num_meshes || $num_results > 0 } {
        GiD_Process Mescape View Zoom Frame escape
    }
    if { $num_results > 0 } {
        #set the first analysis and step to use
        set analysis_to_use [lindex [GiD_Info postprocess get all_analysis] 0]
        set step_to_use [lindex [GiD_Info postprocess get all_steps $analysis_to_use] 0]
        GiD_Process Mescape results analysissel $analysis_to_use $step_to_use escape
    }
    GidUtils::SetWarnLine [_ "Read %d meshes and %d results" $num_meshes $num_results]
    GiD_RaiseEvent GiD_Event_AfterChangeMesh [GiD_Info Mesh -post NumNodes] [GiD_Info Mesh -post NumElements]
    GiD_RaiseEvent GiD_Event_BeforeUpdateWindow SETS
    return
}

#to drag and drop a file
proc Calculix::ReadPostSingleFile { filename } {
    variable _merge
    if { !$_merge } {
        GiD_Process Mescape Files New Yes escape ;#to avoid previous geometry
    }
    GidUtils::EvalAndEnterInBatch Calculix::ReadPost [list $filename]
}


#to add more options to the standar file browser
proc Calculix::MoreImportOptions { f } {
    variable _merge
    ttk::checkbutton $f.chkGeometry -text [_ "Merge"] -variable Calculix::_merge
    grid columnconfigure $f 1 -weight 1
    grid $f.chkGeometry -sticky w -columnspan 2

    variable export_dir
    if { ![info exists export_dir] } {
        #trick to show the more options
        set export_dir open
    }
    return Calculix::export_dir
}

proc Calculix::ReadPostW { } {
    variable _merge
    set multiple 1
    set filenames [MessageBoxGetFilename file read [_ "Read Calculix results file"] \
        {} [list [list {Calculix} {.frd }] [list [_ "All files"] {.*}]] {} $multiple \
        [list [_ "Import options"] Calculix::MoreImportOptions]]
    if {$filenames == ""} {
        return
    }
    if { !$_merge } {
        GiD_Process Mescape Files New Yes escape ;#to avoid previous geometry
    }
    GidUtils::EvalAndEnterInBatch Calculix::ReadPost $filenames
}


#########################################################################################
######################################### OTHER #########################################
#########################################################################################

proc Calculix::AddToMenuPRE { } {
    if { [GidUtils::IsTkDisabled] } {
        return
    }
    if { [GiDMenu::GetOptionIndex Files [list Import "Calculix..."] PRE] != -1 } {
        return
    }
    #try to insert this menu after the word "Files->Import->XYZ points"
    set position [GiDMenu::GetOptionIndex Files [list Import "XYZ points"] PRE]
    if { $position == -1 } {
        set position end
    }
    GiDMenu::InsertOption Files [list Import "Calculix..."] $position PRE Calculix::ReadPreW "" "" insertafter _
}

proc Calculix::AddToMenuPOST { } {
    if { [GidUtils::IsTkDisabled] } {
        return
    }
    if { [GiDMenu::GetOptionIndex Files [list Import "Calculix..."] POST] != -1 } {
        return
    }
    #try to insert this menu after the word "Files->Import->XYZ nodes"
    set position [GiDMenu::GetOptionIndex Files [list Import "XYZ nodes"] POST]
    if { $position == -1 } {
        set position end
    }
    GiDMenu::InsertOption Files [list Import "Calculix..."] $position POST Calculix::ReadPostW "" "" insertafter _
}

proc Calculix::AddToMenu { } {
    Calculix::AddToMenuPRE
    Calculix::AddToMenuPOST
}


Calculix::Init

#invoke this menus changes
Calculix::AddToMenu
GiDMenu::UpdateMenus

#register the proc to be automatically called when re-creating all menus (e.g. when doing files new)
GiD_RegisterPluginAddedMenuProc Calculix::AddToMenu

#register the proc to be automatically called when dropping a file
GiD_RegisterExtensionProc ".frd" POST Calculix::ReadPostSingleFile
#register the proc to be automatically called when dropping a file
GiD_RegisterExtensionProc ".fbd" PRE Calculix::ReadPreSingleFile

#Calculix::ReadPost0 {{C:\Program Files\bConverged\CalculiX\ccx\examples\acou1.frd}}
