#Linux / Mac OS X
#generic implementation using exec of unix system commands        

namespace eval gid_cross_platform {
    variable cache
}

########## PROCESS #####################

proc gid_cross_platform::end_process { pid } {
    set ret_v ""
    # killing twice the same process causes an error
    if { $::tcl_platform(os) == "Darwin" } {
        # in macOS if process is blocked, better use kill -9 instead of kill only
        # kill $pid only sends SIG_TERM to process, but does not kill it
        set err [ catch {
            set ret_v [ exec kill -9 $pid]
        } err_txt]
    } else {
        set err [ catch {
            set ret_v [ exec kill $pid]
        } err_txt]
    }
    return $ret_v
}

proc gid_cross_platform::process_exists { pid } {
    #return [ expr ![ catch {exec ps $pid} ret]]
    # file exists /proc/$pid does not work
    # because when the process ends it's still alive
    # until the parent retrieves it's return code status
    # i.e. it remains <defunct> or zombie
    # that's why we need to use 'ps'
    #return [ file exists /proc/$pid]
    set exists 1
    set err [ catch {exec ps $pid} err_txt]    
    if { $err } {
        set exists 0
        if { [ string first $pid $err_txt] != -1} {
            set exists 1
        }
    }
    return $exists
}

proc gid_cross_platform::get_process_parent { pid } {
    set ppid ""
    set err [ catch {exec ps -o ppid= $pid} err_txt]
    if { !$err} {
        set ppid [ string trim $err_txt]
    }
    return $ppid
}

proc gid_cross_platform::get_process_childs { pid firstonly } {
    set ppids [ list]
    #is this returning all childs or only the first?
    if { $::tcl_platform(os) == "Darwin" } {
        set err [ catch {exec pgrep -P $pid} err_txt]
    } else {
        # Linux
        set err [ catch {exec ps -o pid= --ppid $pid} err_txt]
    }
    if { !$err} {
        set ppids [ string trim $err_txt]
        if { $firstonly } {            
            set ppids [ lindex $pids 0]
        }
    }
    return $ppids
}

proc gid_cross_platform::get_process_name { pid } {
    if { [ catch {exec ps -p $pid --no-headers -o cmd} ret] } {
        set name ""
    } else {
        set name [ file tail $ret]
    }
    return $name
}

proc gid_cross_platform::get_process_ids { } {
    return [ join [ exec ps -A --no-headers -o pid]]
}

proc gid_cross_platform::get_process_ids_from_name { name } {
    return [ exec pidof $name]
}

proc gid_cross_platform::end_all_process_by_name { name } {
    return [ exec killall $name]
}

#bytes
proc gid_cross_platform::get_process_info_memory { pid } {
    #rss  resident set size, the non-swapped physical memory that a task has used (in kiloBytes).                  
    if { [ catch {exec ps -p $pid --no-headers -o rss} ret] } {
        set workingset 0
    } else {
        set ret [ string trim $ret]
        set workingset [ expr {$ret*1024}]
    }
    return $workingset
}

#bytes
proc gid_cross_platform::get_process_info_memory_peak { pid } {
    #vsz  virtual memory size of the process in KiB (1024-byte units). Device mappings are currently excluded
    if { [ catch {exec ps -p $pid --no-headers -o vsz} ret] } {
        set workingset 0
    } else {
        set ret [ string trim $ret]
        set workingset [ expr {$ret*1024}]
    }
    return $workingset        
}

#seconds
proc gid_cross_platform::get_process_info_elapsedtime { pid } {
    #etime  elapsed time since the process was started, in the form [ [ dd-]hh:]mm:ss.
    if { [ catch {exec ps -p $pid --no-headers -o etime} ret] } {
        set elapsedtime 0
    } else {                    
        set ret [ string trim $ret]               
        lassign [ lreverse [ split $ret :-]] ss mm hh dd
        #remove left zeros to avoid consider as an octal number!!
        foreach variable_name {ss mm hh dd} {
            set $variable_name [ string trimleft [ string trim [ set $variable_name]] 0]        
            if { [ set $variable_name] == "" } {
                set $variable_name 0        
            }
        }
        set elapsedtime [ expr {(($dd*24+$hh)*60+$mm)*60+$ss}]
    }
    return $elapsedtime
}

#seconds
proc gid_cross_platform::get_process_info_cputime { pid } {
    #etime  elapsed time since the process was started, in the form [ [ dd-]hh:]mm:ss.
    if { [ catch {exec ps -p $pid --no-headers -o cputime} ret] } {
        set elapsedtime 0
    } else {
        set ret [ string trim $ret]
        lassign [ lreverse [ split $ret :-]] ss mm hh dd
        #remove left zeros to avoid consider as an octal number!!
        foreach variable_name {ss mm hh dd} {
            set $variable_name [ string trimleft [ string trim [ set $variable_name]] 0]        
            if { [ set $variable_name] == "" } {
                set $variable_name 0
            }
        }
        set elapsedtime [ expr {(($dd*24+$hh)*60+$mm)*60+$ss}]
    }
    return $elapsedtime
}

proc gid_cross_platform::get_run_as_administrator_cmd { } {
    set cmd ""
    set lst_sudoers [ list pkexec gksudo kdesudo gksu kdesu ]
    foreach try_cmd $lst_sudoers {
        # set err [ catch {exec which $try_cmd} ret]
        # if { !$err} {
        #     # adding -- to finish passing arguments to XXsudo
        #     # set cmd [ list $ret --]
	#     set cmd $ret
        #     break
        # }
        set cmd [ auto_execok $try_cmd]
        if { $cmd != ""} {
            break
        }
    }
    if { $cmd == "" } {
        # brute force....
        # set err_xterm [ catch {exec which xterm} full_xterm]
        # set err_sudo [ catch { exec which sudo } full_sudo]
        # if { !$err_xterm && !$err_sudo} {
        #     set cmd [ list $full_xterm -e $full_sudo]
        # }
        set full_xterm [ auto_execok xterm]
        set full_sudo [ auto_execok sudo]
        if { ( $full_xterm != "") && ( $full_sudo != "")} {
            set cmd [ list $full_xterm -e $full_sudo]
        }
    }
    return $cmd
}

proc gid_cross_platform::get_run_as_administrator_cmdline { args} {
    set cmd ""
    if { $::tcl_platform(os) == "Darwin" } {
        # https://stackoverflow.com/questions/1517183/is-there-any-graphical-sudo-for-mac-os-x
        # osascript -e "do shell script \"$*\" with administrator privileges"
        set cmd [ list osascript -e \
                      "do shell script \"$args\" with administrator privileges" \
                     ]
    } else {
        # linux
        set run_as_administrator_cmd [ gid_cross_platform::get_run_as_administrator_cmd]
        set cmd [ list {*}$run_as_administrator_cmd {*}$args]
    }
    return $cmd
}

proc gid_cross_platform::run_as_administrator { args } {
    #ignore optional arguments -show $show, to have the same syntax as Windows
    #set show normal
    if { [lindex $args 0] == "-show" } {
        #set show [lindex $args 1]
        set args [lrange $args 2 end]
    }
    set cmd [ gid_cross_platform::get_run_as_administrator_cmdline {*}$args ]
    set err [ catch {exec {*}$cmd} ret]
    if { $err} {
        error "Executing run_as_administrator_cmd $args failed: $ret"
    }
    return $err
}

#to open the file explorer in a specified folder
proc gid_cross_platform::open_by_extension { url } {
    # open is the OS X equivalent to xdg-open on Linux, start is used on Windows
    if { $::tcl_platform(os) == "Darwin" } {
        set browser open
    } else {
        # Linux
        set browser xdg-open
    }
    set command [ auto_execok $browser]    
    if {[ string length $command] == 0} {
        return -code error "couldn't find browser"
    }
    # make gid dynamic libraries do not affect external programs
    set url [ regsub -all {&} $url {\\\&}]
    if {[ catch {exec sh -c "export LD_LIBRARY_PATH=; $command $url" &} error]} {
        return -code error "couldn't execute '$command': $error"
    }
}

# execute an external program with args
proc gid_cross_platform::execute_program_in_path { program args } {
    # GiD mangles with encodings, so the environment variable PATH in windows is invalid:
    # set err [ catch { set full_program [ exec which $program]} error]    
    set full_program [ auto_execok $program]
    if { $full_program == ""} {
        return -code error "couldn't execute '$program': $error"
    }
    # get first option
    set lst_programs [ split $full_program \n]
    set full_program [ lindex $lst_programs 0]
    set err 0
    set error_txt ""
    set err [ catch {exec "$full_program" {*}$args &} error_txt]    
    if { $err} {
        return -code error "couldn't execute '$program': $error"
    }
}

########## GUI #####################

proc gid_cross_platform::get_desktop_workarea { } {
    if { [ info commands winfo]!= "" && [ winfo exists .] } {
        set w [ winfo screenwidth .] 
        set h [ winfo screenheight .]
    } else {
        set w 0
        set h 0
    }
    return [ list 0 0 $w $h]
}

# Copy the contents of the photo image to clipboard
proc gid_cross_platform::image_to_clipboard { image } {
    #how to handle the clipboard image in Linux / MacOSX ?
    #there are lots of clipboard managers but nothing standard !
    return ""
}

# Copy the contents of the clipboard into a photo image. Return the photo identifier.
proc gid_cross_platform::clipboard_to_image {} {
    #how to handle the clipboard image in Linux / MacOSX ?
    #there are lots of clipboard managers but nothing standard !
    return ""
}

proc get_list_displays_mac_book_pro {} {
    # mac book pro has 2 graphics cards: intel and Radeon
    # depending on which application is running, display (spdisplay_ndrvs) is attache to one or the other
    # Check issue GID-3185 to get the files
    # set fi [ open list_displays_mac_book_pro_intel.txt]
    set fi [ open list_displays_mac_book_pro_radeon.txt]
    set json_data [ read $fi]
    close $fi
    return $json_data
}

# returns a dictionary with key = name of monitor, value = dictionary with info
proc gid_cross_platform::get_display_monitors {} {
    # atm only for macos
    set ret [ dict create]
    if { $::tcl_platform(os) == "Darwin"} {
        set json_data [ exec system_profiler SPDisplaysDataType -json]
        # set json_data [ get_list_displays_mac_book_pro]
        set json_data [ regsub -all {\n} [ regsub -all {\r} $json_data {}] {} ]
        set json_data [ regsub -all {\:\[} $json_data {\: \[}]
        set json_data [ regsub -all {\:\[} $json_data {\: \[}]
        package require gid_json
        # check in 16.1.10d this in macOS as 'gid_json::json2dict' has changed, it returns dict without \"\"
        set dict_data [ lindex [ gid_json::json2dict $json_data] 0]
        # may be we have several graphics cards attached, like mac book pro
        # choose the card with Display attached, i.e. which contains spdisplays_ndrvs
        set displays_data ""
        foreach card_data [ dict get $dict_data SPDisplaysDataType] {
            if { [ dict exists $card_data spdisplays_ndrvs]} {
                # W "found displays connected to [ dict get $card_data _name]"
                set displays_data $card_data
                break
            }
        }
        if { $displays_data == ""} {
            # W "No displays found"
            return $ret
        }
        # set displays_data [ lindex [ dict get $dict_data SPDisplaysDataType] 0]
        set list_monitors [ dict get $displays_data spdisplays_ndrvs]
        set dict_monitors_info [ dict create]
        foreach mon_info $list_monitors {
            set dict_mon [ dict create]
            foreach [ list k v ] $mon_info {
                set dict_key [ regsub -all {_} $k {-}]
                dict set dict_mon $dict_key $v
            }
            if { [ dict exists $dict_mon -name]} {

                # calculate scaling
                if { [ dict exists $dict_mon -spdisplays-pixels] && \
                         [ dict exists $dict_mon -spdisplays-resolution]} {
                    set disp_pixels [ dict get $dict_mon -spdisplays-pixels]
                    set disp_resol [ dict get $dict_mon -spdisplays-resolution]
                    set exp_w_h {^([0-9]+) *x *([0-9]+)$}
                    set exp_w_h_hz {^([0-9]+) *x *([0-9]+) *@ *[0-9]+ *Hz$}
                    if { [ regexp $exp_w_h $disp_pixels dum wp hp] == 1} {
                        # -spdisplays-resolution can be
                        # "1920 x 1080" or "1920 x 1080 @ 60 Hz"
                        if { ( [ regexp $exp_w_h $disp_resol dum wr hr] == 1) || \
                                 ( [ regexp $exp_w_h_hz $disp_resol dum wr hr] == 1)} {
                            # scaling = width_pixels / width_resolution
                            set scaling [ expr $wp / $wr]
                            dict set dict_mon -scale-factor $scaling
                        }
                    }
                }

                # storing monitor info in global dictionary
                set mon_name [ dict get $dict_mon -name]
                dict set dict_monitors_info $mon_name $dict_mon
            }
        }
        set ret $dict_monitors_info
    }
}

########## PLATFORM #####################

proc gid_cross_platform::getUsername {} {
    if { [ catch {
        return $::env(USER)
    }]} {
        return [ exec whoami]
    }
}

proc gid_cross_platform::getHostname {} {
    set hostname ""
    if { $::tcl_platform(os) == "Darwin"} {
        if { [ info exist ::env(HOST)]} {
            set hostname $::env(HOST)
        }
    } else {
        if { [ info exist ::env(HOSTNAME)]} {
            set hostname $::env(HOSTNAME)
        }
    }
    if { $hostname == ""} {
        catch {
            set hostname [ exec hostname]
        }
    }
    return $hostname
}

#return one of:  "Windows 32" "Windows 64" "Linux 32" "Linux 64" "MacOSX 32" "MacOSX 64" "$::tcl_platform(os) ?"
proc gid_cross_platform::get_current_platform {} {
    if { $::tcl_platform(os) == "Linux" } {
        if { $::tcl_platform(machine) == "x86_64" } {
            set myplatform "Linux 64"
        } else {
            set myplatform "Linux 32"
        }
    } elseif { $::tcl_platform(os) == "Darwin" } {
        if { $::tcl_platform(machine) == "Power Macintosh" || $::tcl_platform(machine) == "i386" } {
            set myplatform "MacOSX 32"
        } else {
            set myplatform "MacOSX 64"
        }
    } else {
        #AIX , HP-UX, IRIX, SunOS, DragonFly, FreeBSD, NetBSD, OpenBSD, OSF1
        set myplatform "$::tcl_platform(os) ?"
    }
    return $myplatform
}

########## NETWORK #####################

#Get the servers ip address (from http://purl.org/mini/tcl/526.html)
#try to avoid this,it is too slow  
proc gid_cross_platform::get_ip { } {
    if { [ catch {  
        set me [ socket -server xxx -myaddr [ info hostname] 0]
        set ip [ lindex [ fconfigure $me -sockname] 0]
        close $me
    } err] } {
        set ip 0.0.0.0
    }
    return $ip
}

########## FILE SYSTEM #####################

proc gid_cross_platform::get_path_separator { } {
    return ":"
}

#return the temporary folder
proc gid_cross_platform::get_tmp { } {
    return "/tmp"
}

#by now AWS appstream offer only Windows platform
proc gid_cross_platform::is_appstream { } {
    return 0
}

# specials folder in unix and macOs
# name: home downloads documents desktop publicshare
proc gid_cross_platform::get_folder_special { name } {
    variable cache
    if  { ![ info exists cache_folders($name)] } {                         
        set dir ""
        if { $::tcl_platform(os) == "Darwin"} {
            array set registry_name [ list \
                                          home "home folder" downloads "downloads folder" \
                                          documents "documents folder" desktop "desktop" \
                                          publicshare "public folder" \
                                         ]
            if { [ info exists registry_name($name)] } {
                set apple_script "
do shell script \"echo \" & ( POSIX path of ( path to $registry_name($name) ) )
"                    
                catch { 
                    set dir [ exec echo $apple_script | osascript] 
                }
            }
        } else {
            array set registry_name [ list \
                                          home "HOME" downloads "DOWNLOAD" \
                                          documents "DOCUMENTS" desktop "DESKTOP" \
                                          publicshare "PUBLICSHARE" \
                                          ]
            if { [ info exists registry_name($name)] } {
                catch { 
                    set dir [ exec xdg-user-dir $registry_name($name)]
                }
            }
        }
        set cache(folders,$name) [ file normalize $dir]        
    }
    return $cache(folders,$name)
}

########## COMPRESSION #####################

proc gid_cross_platform::get_unzip_exe { } {
    return unzip
}

#e.g.
#gid_cross_platform::unzip "-o -u -a" "/tmp/my_file.zip" "/tmp/out" 0
proc gid_cross_platform::unzip { flags zipfile {extraction_dir ""} {run_as_administrator 0}} {
    set exe [ gid_cross_platform::get_unzip_exe]
    set command "$exe {*}$flags \"$zipfile\" -d \"$extraction_dir\""
    if { $run_as_administrator } {
        set command [ gid_cross_platform::get_run_as_administrator_cmdline {*}$command]
    }
    exec $command
}

proc gid_cross_platform::get_zip_exe { } {    
    return zip
}

#e.g.
#gid_cross_platform::zip "-r -9" "C:/my_file.zip" "C:/temp/my_folder_to_add" 0
proc gid_cross_platform::zip { flags zipfile {file_to_add ""} {run_as_administrator 0} } {
    set exe [ gid_cross_platform::get_zip_exe]
    set dir_prev [ pwd]
    set file_to_add_dir [ file dirname $file_to_add]
    set file_to_add_tail [ file tail $file_to_add]
    set command "$exe $flags \"$zipfile\" \"$file_to_add_tail\""
    cd $file_to_add_dir
    if { $run_as_administrator } {
        set command [ gid_cross_platform::get_run_as_administrator_cmdline {*}$command]
    }
    exec {*}$command
    cd $dir_prev
}

########## LINUX SERVICES (DAEMON) #####################
#e.g.
# set fullname "C:/Program Files/PasServer 2.7/passerverd"
# set service_description "GiD password server"
# set service_port 7073
# set res [gid_cross_platform::create_daemon $fullname $service_description $service_port 0]
# if { [lindex $res 0] == "1" } {
#     error [lindex $res 1]
# }
proc gid_cross_platform::create_daemon { progname progtitle port { uninstall 0 } } {
    set res [gid_cross_platform::create_daemon_systemd $progname $progtitle $port $uninstall]
    #if { [lindex $res 0] == "1" } {
    #    set res [gid_cross_platform::create_daemon_rclocal $progname $progtitle $port $uninstall]
    #    if { [lindex $res 0] == "1" } {
    #        set res [gid_cross_platform::create_daemon_initd $progname $progtitle $port $uninstall]
    #    }
    #}
    return $res
}

proc gid_cross_platform::delete_daemon { service_name  } {
    return [gid_cross_platform::create_daemon $service_name "" "" 1]
}

#service_name is implicit in progname
proc gid_cross_platform::create_daemon_systemd { progname progtitle port { uninstall 0 } } {
    set msg "" 
    set res [list 1 $msg]
    set service_unit_folder /etc/systemd/system
    set service_name [file rootname [file tail $progname]]
    if { $uninstall } {
        if { [catch {exec systemctl stop ${service_name}.service} msg] } {
            set res [list 1 $msg]
        }
    }
    if { [file exists $service_unit_folder] && [file isdirectory $service_unit_folder] } {
        set service_unit_filename [file join $service_unit_folder ${service_name}.service]       
        if { $uninstall } {
            if { [file exists $service_unit_filename] && [file isfile $service_unit_filename] } {
                if { [catch { file delete $service_unit_filename } err] } {
                    set  res [list 1 "You do not have enough permissions to delete the file $service_unit_filename. Try again as 'root'"]
                } else {
                    set msg "$progname has been correctly uninstalled"
                    set res [list 0 $msg]
                }
            } else {
                set  res [list 1 "File $service_unit_filename not found"]
            }         
        } else {
            if { [catch { set fout [open $service_unit_filename w]} err] } {
                set  res [list 1 "You do not have enough permissions to write the file $service_unit_filename. Try again as 'root'"]
            } else {
                puts $fout {[Unit]}
                puts $fout "Description=$progtitle: floating password for GiD and problemtypes"
                puts $fout {Requires=network.target}
                puts $fout {After=systemd-user-sessions.service}
                puts $fout {StartLimitIntervalSec=0}
                puts $fout {}
                puts $fout {[Service]}
                puts $fout {Type=simple}
                puts $fout "ExecStart=$progname -port $port"
                puts $fout {Restart=always}
                puts $fout {RestartSec=1}
                puts $fout {}
                puts $fout {[Install]}
                puts $fout {WantedBy=multi-user.target}
                close $fout
                set msg "$progname has been correctly installed in port $port." 
                append msg "\n" "It will be executed every time the computer restarts."
                set res [list 0 $msg]
            }
        }
    } else {
        set res [list 1 "$service_unit_folder folder not found"]
    }
    if { !$uninstall&& [lindex $res 0]=="0" } {
        if { [catch {exec systemctl enable ${service_name}.service} msg] } {
            set res [list 1 $msg]
        }
        if { [catch {exec systemctl start ${service_name}.service} msg] } {
            set res [list 1 $msg]
        }
    }
    return $res
}

proc gid_cross_platform::create_daemon_rclocal { progname progtitle port { uninstall 0 } } {
    if { [file pathtype $progname] == "relative" } {
        return [list 1 "$progname cannot be a relative path"]
    }
    set localf ""
    foreach f {/etc/rc.d/rc.local /etc/rc.local} {
        if { [file exists $f] && [file isfile $f] } {
            set localf $f
            break
        }
    }
    if { $f == "" } {
        return [list 1 "$progname cannot be un/installed as daemon in this computer"]
    }
    if { ![file writable $localf] } {
        return [list 1 "You do not have enough permissions to perform the un/installation. \
                Perform this step again as 'root'"]
    }
    set localfback [file join [file dirname $localf] [file tail $localf].back]
    file copy -force $localf $localfback
    file attributes $localfback -permissions 0644
    
    if { [catch {
        set fin [open $localfback r]
    } err] } {
        return [list 1 "You do not have enough permissions to perform the un/installation. \
                Perform this step again as 'root'"]
    }
    if { [catch {
        set fout [open $localf w]
    } err] } {
        return [list 1 "You do not have enough permissions to perform the un/installation \
                Perform this step again as 'root'"]
    }
    set found 0
    while { ![eof $fin] } {
        gets $fin aa
        if { [string match "*$progname*" $aa] } { 
            set found 1
            continue
        }
        if { [eof $fin] && $aa == "" } { break }
        puts $fout $aa
    }
    if { $uninstall == 0 } {
        puts $fout "# $progname $progtitle"
        puts $fout "$progname -port $port &"
    }
    close $fin
    close $fout
    file attributes $localf -permissions 0777
    set msg "The following file has been modified: $localf."
    append msg "\n" "A backup copy of '$localf' has been saved as '$localfback'"
    if { $uninstall == 0 } {
        #exec chmod +s $progname
        file attributes $progname -permissions +s
        append msg "\n" "$progname has been correctly installed using port $port." 
        append msg "\n" "It will be executed every time the computer restarts."
    } else {
        if { !$found } {
            return [list 1 "$progname not found to be uninstalled."]
        } else {
            append msg "\n" "$progname has been correctly uninstalled." 
            append msg "\n" "It will not execute anymore when computer restarts."
        }
    }
    return [list 0 $msg]
}

proc gid_cross_platform::create_daemon_initd { progname progtitle port { uninstall 0 } } {
    if { [file pathtype $progname] == "relative" } {
        return [list 1 "$progname cannot be a relative path"]
    }
    set msg ""    
    set fdir "/etc/init.d"
    if { ![file exists $fdir] || ![file isdirectory $fdir] } {
        return [list 1 "$progname cannot be un/installed as daemon in this computer"]
    }
    if { ![file writable $fdir] } {
        return [list 1 "You do not have enough permissions to perform the un/installation. \
                Perform this step again as 'root'"]
    }
    set fdirlink ""
    foreach f {/etc/rc2.d /etc/rc3.d /etc/rc.d/rc2.d /etc/rc.d/rc3.d} {
        if { [file exists $f] && [file isdirectory $f]} {
            set fdirlink $f
            break
        }
    }
    if { $fdirlink == "" } {
        return [list 1 "$progname cannot be un/installed as daemon in this computer"]
    }
    if { ![file writable $fdirlink] } {
        return [list 1 "You do not have enough permissions to perform the un/installation. \
                Perform this step again as 'root'"]
    }
    set fscript [file join $fdir [file tail $progname]]
    set link [file join $fdirlink S99[file tail $progname]]
    if { $uninstall == 0 } {
        #install
        if { [catch {
            set fout [open $fscript w]
        } err] } {
            return [list 1 "You do not have enough permissions to perform the installation. \
                    Perform this step again as 'root'"]
        }
        set contents [join [list "#!/bin/sh\n"\
                "#init script for $progname\n"\
                "PIDFILE=/var/run/[file tail $progname].pid\n"\
                "case \"\$1\" in\n"\
                "  'start')\n"\
                "        echo -n  \"Starting [file tail $progname]...\"\n"\
                "        $progname -port $port &\n"\
                "        echo \$! > \$PIDFILE\n"\
                "        echo \"done.\"\n"\
                "        ;;\n"\
                "  'stop')\n"\
                "       kill `cat \$PIDFILE`\n"\
                "       rm -f \$PIDFILE\n"\
                "        ;;\n"\
                "  *)\n"\
                "        echo \"usage: \$0 {start|stop}\"\n"\
                "        ;;\n"\
                "esac\n"\
                "exit 0\n"] ""]
        puts $fout $contents
        close $fout
        file attributes $fscript -permissions 0711
        #exec ln -s $fscript $link
        #exec chmod +s $progname
        file link -symbolic $link $fscript
        file attributes $progname -permissions +s
        
        set msg "$progname has been correctly installed in port $port." 
        append msg "\n" "It will be executed every time the computer restarts."
    } else {
        #uninstall
        if { ![file exists $link] } {
            return [list 1 "$progname not found"]
        } else {
            file delete $link
            file delete $fscript 
        }
    }
    return [list 0 $msg]
}

proc gid_cross_platform::start_service { service_name } {
    set fail 0
    set msg ""
    if { [catch {exec systemctl start ${service_name}.service} msg] } {
        set fail
    }
    return [list $fail $msg]
}

proc gid_cross_platform::stop_service { service_name } {
    set fail 0
    set msg ""
    if { [catch {exec systemctl stop ${service_name}.service} msg] } {
        set fail
    }
    return [list $fail $msg]
}

########## OTHER #####################

